# M4 FORTH benchmarks

## Contents

- [Contents](#contents)
- [Testing](#testing)
  - [Factorization](#factorization)
  - [Gcd 1](#gcd-1)
  - [Gcd 2](#gcd-2)
  - [Fib 1](#fib-1)
  - [Fib 2](#fib-2)
  - [Forth Nesting Benchmark](#forth-nesting-benchmark)
  - [Forth Sieve Benchmark](#forth-sieve-benchmark)
  - [FILLIN](#fillin)
  - [Project Euler 21: Amicable numbers](#project-euler-21-amicable-numbers)
  - [N-body problem](#n-body-problem)
  - [Pangram Benchmark](#pangram-benchmark)
    - [Forth from Rosettacode](#forth-from-rosettacode)
    - [Double BITSET version](#double-bitset-version)
    - [Double BITSET inline version](#double-bitset-inline-version)
    - [C version](#c-version)
 - [Logical conditions with 8, 16 and 32 bit constants](#logical-conditions-with-8-16-and-32-bit-constants)
 

## Testing

There are examples of small benchmarks on the https://theultimatebenchmark.org/ website. Including speeds on individual platforms.

[Back to contents](#contents)

### Factorization

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/decomp.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/decomp.asm

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/decomp_fast.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/decomp_fast.asm

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/decomp.c

|       Name        |                System                |        Forth / C               |  Benchmark   | Time (sec/round) |
| :---------------: | :----------------------------------: | :----------------------------: | :----------: | :--------------- |
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH                       | Decompose    | 1m6.81s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH div:old               | Decompose    | 1m0.03s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH div:old_fast          | Decompose    | 47.54s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH div:fast              | Decompose    | 1m3.07s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH mul:fast              | Decompose    | 1m3.07s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH mul:old_fast          | Decompose    | 1m4.75s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH mul:fast div:old_fast | Decompose    | 43.08s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | zcc z88dk v16209               | Decompose    | 1m40.47s

[Back to contents](#contents)

### Gcd 1

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/gcd1.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/gcd1.asm

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/boriel_gcd1.bas
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/boriel_gcd1.asm

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/gcd1.c

|        Name       |                  System                   |       Forth / C         | Benchmark | Time (sec/round) | Scale |
| :---------------: | :---------------------------------------: | :---------------------: | :-------: | :--------------- | :---: |
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu             | M4_FORTH                |    GCD1   | 3.73s
| Dw0rkin           | ZX Spectrum Fuse 1.6.0 Ubuntu             | Boriel Basic zxbc 1.16.4|    GCD1   | 11.27s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu             | zcc z88dk v16209        |    GCD1   | 31.72s
| Ben               | IBM PS/2 L40SX                            | DX-Forth                |    GCD1   | 0m29s            | 10x
| Thunder.Bird      | Amstrad PPC 512                           | DX-Forth                |    GCD1   | 0m48s            | 5x
| Alexander Muller  | IBM PS/2 L40SX                            | DX-Forth                |    GCD1   | 0m15s            | 5x
| Alexander Muller  | Raspberry Pi ARM 700Mhz                   | Gforth 0.7.0            |    GCD1   | 0m04s            | 100x
| Thorsten Schoeler | Sinclair Spectrum+                        | Aber Forth (FIG-Forth)  |    GCD1   | 2m14s
| Wolfgang Stief    | SUN SparcStation 10 TI TMS390255          | OpenFirmware            |    GCD1   | 0,51s
| Wolfgang Stief    | SUN Ultra 1 200 Mhz UltraSprac            | OpenBoot 3.25           |    GCD1   | 0,08s
| Stefan Niestegge  | Atari Falcon 68060                        | f68kans                 |    GCD1   | 0,063s
| Herbert Lange     | Compaq Deskpro P166                       | pForth V27              |    GCD1   | 0,002s
| Herbert Lange     | Apple iMac G3 400Mhz                      | pForth V27              |    GCD1   | 0,063s
| Herbert Lange     | DEC 3000 400s                             | pForth V27              |    GCD1   | 0,483
| Herbert Lange     | SUN Ultra 1 Creator 3D                    | pForth V27              |    GCD1   | 0,022s
| Johan Kotlinski   | C64                                       | DurexForth 1.6.1 (STC)  |    GCD1   | 60s               | 1x
| Carsten Fulde     | Amiga 500 mit ACA1233n-Turbokarte (68030 @ 40 MHz) | JForth 3.1 Delta Research (Phil Burk) | GCD1 | 0.4s | 1x
| Carsten Fulde     | Enterprise 128                            | IS-Forth von B. Tanner  |    GCD1   | 78s     | 1x
| Carsten Fulde     | Sinclair QL M68008 7,5 Mhz                | Forth79 (G.W.Jackson)   |    GCD1   | 21
| Carsten Strotmann | Nixdorf 8810 M15 (8Mhz i286)              | L.O.V.E. Forth          |    GCD1   | 6
| Enrico/Dirk       | Robotron A 7150 i8086/8087 Multibus ~5Mhz | VolksForth MS-DOS (ITC) |    GCD1   | 25
| Ralf Neumann      | Yodabashi Formula 1 Z80 4Mhz              | VolksForth CP/M (ITC)   |    GCD1   | 42
| Carsten Strotmann | Scheider Tower AT 220 i286 10Mhz          | VolksForth MS-DOS (ITC) |    GCD1   | 5
| Martin Metz       | Amstrad NC100 Z80 4.606Mhz                | VolksForth CP/M (ITC)   |    GCD1   | 38.1
| Andreas Boehm     | Commodore C64 6510                        | Audiogenic Forth-64     |    GCD1   | 215.52
| Thorsten Kuphaldt | Amiga 3000 68030 25Mhz                    | jforth                  |    GCD1   | 0.64
| Ingo Soetebier    | iBook PPC 750lx (G3) 600Mhz               | OpenFirmware            |    GCD1   | 0.024
| Matthias Trute    | Atmega16 8MHz                             | amForth 4.4             |    GCD1   | 7.12
| Norbert Kehrer    | Mupid II (BTX Decoder)                    | FIG-Forth 1.1           |    GCD1   | 205s
| Norbert Kehrer    | Mupid II (BTX Decoder)                    | Camel Forth 1.01        |    GCD1   | 116s

[Back to contents](#contents)

### Gcd 2

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/gcd2.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/gcd2.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/gcd2u.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/gcd2u.asm

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/gcd2.c

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/boriel_gcd2.bas
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/boriel_gcd2.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/boriel_gcd2u.bas
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/boriel_gcd2u.asm

I had to manually delete lines starting with:

    #line

Rename labels:

    __BOR16: -->  .core.__BOR16:
    __EQ16:  -->  .core.__EQ16:
    etc.

Erase the lines:

    push namespace core
    pop namespace


|       Name        |                  System                   |          Forth / C        | Benchmark | Time (sec/round) | Scale |
| :---------------: | :---------------------------------------: | :-----------------------: | :-------: | :--------------- | :---: |
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu             | M4_FORTH                  |   GCD2    | 4.50s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu             | M4_FORTH                  |   GCD2 ui | 4.03s
| Dw0rkin           | ZX Spectrum Fuse 1.6.0 Ubuntu             | Boriel Basic zxbc 1.16.4  |   GCD2    | 12.52s
| Dw0rkin           | ZX Spectrum Fuse 1.6.0 Ubuntu             | Boriel Basic zxbc 1.16.4  |   GCD2 ui | 10.36s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu             | zcc z88dk v16209          |   GCD2    | 35.31s
| Ben               | IBM PS/2 L40SX                            | DX-Forth                  |   GCD2    | 42s              | 10x
| Wolfgang Stief    | SUN SparcStation 10 TI TMS390255          | OpenFirmware              |   GCD2    | 0,65s            |
| Wolfgang Stief    | SUN Ultra 1 200 Mhz UltraSprac            | OpenBoot 3.25             |   GCD2    | 0,11s            |
| Stefan Niestegge  | Atari Falcon 68060                        | f68kans                   |   GCD2    | 0,067s           |
| Martin Neitzel    | Asus EeePC 1000h (Atom N270 1.6Ghz)       | FreeBSD 9 FICL Bootloader |   GCD2    | 0.57s            |
| Johan Kotlinski   | C64                                       | DurexForth 1.6.1 (STC)    |   GCD2    | 70s              | 1x
| Carsten Strotmann | A-ONE (Apple 1 Clone) mit 65C02           | TaliForth 2 (STC)         |   GCD2    | 1m25s            | 1x
| Thomas Woinke     | Steckschwein 8MHz 65c02                   | TaliForth 2 (STC)         |   GCD2    | 11.75s           | 1x
| Enrico/Dirk       | Robotron A 7150 i8086/8087 Multibus ~5Mhz | VolksForth MS-DOS (ITC)   |   GCD2    | 30               |
| Andreas Boehm     | Commodore C64 6510                        | Audiogenic Forth-64       |   GCD2    | 84.84            |
| Matthias Trute    | Atmega16 8MHz                             | amForth 4.4               |   GCD2    | 10.5             |
| Norbert Kehrer    | Mupid II (BTX Decoder)                    | FIG-Forth 1.1             |   GCD2    | 188s             |
| Norbert Kehrer    | Mupid II (BTX Decoder)                    | Camel Forth 1.01          |   GCD2    | 135s             |

[Back to contents](#contents)

### Fib 1

Warning: The code for FORTH and C is adopted and contains an error where it correctly calculates only for positive values.
Nevertheless, in both variants, a signed integer is calculated and no one is favored to use an unsigned integer.
Furthermore, it actually calculates fib(n+1) instead of fib(n). So it returns 1 for negative numbers and zero.

    if (n<2)
        return 1;
    else
        return fib(n-1)+fib(n-2);

    DUP PUSH(2) LT IF
        DROP PUSH(1) SEXIT
    THEN
    DUP  _1SUB SCALL(fib1s)
    SWAP _2SUB SCALL(fib1s) ADD

It should correctly return from the definition for:

    Fib(0) = 0
    Fib(1) = 1
    Fib(n) = Fib(n-1) + Fib(n-2)

    if (n<2)
        return n;
    else
        return fib(n-1)+fib(n-2);

    DUP PUSH(2) LT IF
        SEXIT
    THEN
    DUP  _1SUB SCALL(fib1s)
    SWAP _2SUB SCALL(fib1s) ADD

Or a faster but incorrect variant for Fib(0) is:

    Fib(0) = 0 ... fail
    Fib(1) = 1
    Fib(2) = 1
    Fib(n) = Fib(n-1) + Fib(n-2)

    if (n<3)
        return 1;
    return fib(n-1)+fib(n-2);

    DUP PUSH(3) LT IF
        SEXIT
    THEN
    DUP  _1SUB SCALL(fib1s)
    SWAP _2SUB SCALL(fib1s) ADD

or

    DUP PUSH(2) GT IF
        DUP  _1SUB SCALL(fib1s)
        SWAP _2SUB SCALL(fib1s) ADD
        SEXIT
    THEN
    DROP PUSH(1)

And negative values should be counted as positive and the result should be turned to a negative value.
The algorithm could be greatly speeded up using CASE, but it would have to be used for both C and FORTH to make the results comparable.

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/fib1.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/fib1.asm

Use data stack for function:

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/fib1s.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/fib1s.asm

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/fib1.c


|       Name        |               System                 |        Forth / C        |  Benchmark  | Time (sec/round) | Scale |
| :---------------: | :----------------------------------: | :---------------------: | :---------: | :--------------- | :---: |
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH                | Fib1        | 30m26.87s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH use data stack | Fib1s       | 16m35.63s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | zcc z88dk v16209        | Fib1        | 41m50.97s
| Johan Kotlinski   | DEC 3000-600 Alpha 21064 175Mhz      | pForth                  | Fibonacci 1 | 0.0038
| Ingo Soetebier    | iBook PPC 750lx (G3) 600Mhz          | OpenFirmware            | Fibonacci 1 | 0.0026
| Michael Kalus     | Rockwell R1200-14, 2Mhz 65F12        | RSC-Forth               | Fibonacci 1 | 16.09
| Matthias Trute    | Atmega16 8MHz                        | amForth 4.4             | Fibonacci 1 | 1.46
| Carsten Strotmann | IBM L40X (386SX)                     | VolksForth MS-DOS       | Fibonacci 1 | 0.36s
| Wolfgang Stief    | SUN SparcStation 10 TI TMS390255     | OpenFirmware            | Fib1        | 0,005s
| Wolfgang Stief    | SUN Ultra 1 200 Mhz UltraSprac       | OpenBoot 3.25           | Fib1        | 0,014s
| Herbert Lange     | Compaq Deskpro P166                  | pForth V27              | Fib1        | 0,061s
| Herbert Lange     | Apple iMac G3 400Mhz                 | pForth V27              | Fib1        | 0,015s
| Herbert Lange     | DEC 3000 400s                        | pForth V27              | Fib1        | 0,098s
| Herbert Lange     | SUN Ultra 1 Creator 3D               | pForth V27              | Fib1        | 0,052s
| Thorsten Schoeler | NCR 3150 486SX/25Mhz+FPU=Linux 2.0.0 | gforth 0.3.0            | Fib1        | 1m79s
| Thorsten Schoeler | MSP430FR5739, 8Mhz DCO intern MSP-EXP430FR5739 Experimenter Board |CamelForth|FIB1| 00'46':39 | 100x

[Back to contents](#contents)

### Fib 2

Wariant with `?do`. Because the variant with `do` starts `0 0 do` and executes 65536!

    : fib2 ( n1 -- n2 )
        0 1 ROT 0 ?DO
            OVER_ADD SWAP LOOP
        DROP
    ;

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/fib2.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/fib2.asm

Use data stack for function, do loop --> for next:

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/fib2s.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/fib2s.asm

To get an idea of how far the resulting code is from one of the best. Handwritten M4 word in assembler replacing C function "fib2":

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/fib2a.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/fib2a.asm

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/fib2.c

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/boriel_fib2.bas
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/boriel_fib2.asm


|       Name        |               System                 |        Forth / C        |  Benchmark  | Time (sec/round) | Scale |
| :---------------: | :----------------------------------: | :---------------------: | :---------: | :--------------- | :---: |
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH                | Fib2        | 0m5.65s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH use data stack | Fib2s       | 0m5.03s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH use assembler  | Fib2a       | 0m2.55s
| Dw0rkin           | ZX Spectrum Fuse 1.6.0 Ubuntu        | Boriel Basic zxbc 1.16.4| Fib2 a = a+c| 0m14.38s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | zcc z88dk v16209        | Fib2 a = a+c| 0m49.19s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | zcc z88dk v16209        | Fib2 a+= c  | 0m43.97s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | zcc z88dk v19766 -O2    | Fib2 a+= c  | 0m36.09s
| Carsten Strotmann | Scheider Tower AT 220 i286 10Mhz     | VolksForth MS-DOS (ITC) | Fibonacci2  | 8
| Johan Kotlinski   | C64                                  | DurexForth 1.6.1 (STC)  | Fibonacci 2 | 1m57s   | 1x
| Carsten Fulde     | Amiga 500 mit ACA1233n-Turbokarte (68030 @ 40 MHz) | JForth 3.1 Delta Research (Phil Burk) | Fibonacci 2 | 0.35s | 1x
| Carsten Fulde     | Enterprise 128                       | IS-Forth von B. Tanner  | Fibonacci 2 | 118.4s  | 1x
| Carsten Strotmann | A-ONE (Apple 1 Clone) mit 65C02      | TaliForth 2 (STC)       | Fibonacci 2 | 1m50s   | 1x
| Thomas Woinke     | Steckschwein 8MHz 65c02              | TaliForth 2 (STC)       | Fibonacci 2 | 15.23s  | 1x
| Carsten Fulde     | Sinclair QL M68008 7,5 Mhz           | Forth79 (G.W.Jackson)   | Fibonacci 2 | 35      |
| Enrico/Dirk       |Robotron A 7150 i8086/8087 Multibus ~5Mhz|VolksForth MS-DOS (ITC)| Fibonacci 2| 46      |
| J. Kunz           | DEC 3000-600 Alpha 21064 175Mhz      | pForth                  | Fibonacci 2 | 0.00001425 |
| Stefan Herold     | Amstrad 6128+ Z80A 4Mhz              | Uniforth                | Fibonacci 2 | 0.23    |
| Ingo Soetebier    | iBook PPC 750lx (G3) 600Mhz          | OpenFirmware            | Fibonacci 2 | 0.0027  |
| Michael Kalus     | Rockwell R1200-14, 2Mhz 65F12        | RSC-Forth               | Fibonacci 2 | 0.05    |
| Matthias Trute    | Atmega16 8MHz                        | amForth 4.4             | Fibonacci 2 | 0.0047  |
| Carsten Strotmann | Nixdorf 8810 M15 (8Mhz i286)         | L.O.V.E. Forth          | Fibonicci 2 | 8
| Ben               | IBM PS/2 L40SX                       | DX-Forth                | Fib2        | 1m03s   | 10x
| Thunder.Bird      | Amstrad PPC 512                      | DX-Forth                | Fib2        | 1m45s   | 5x
| Alexander Muller  | IBM PS/2 L40SX                       | DX-Forth                | fib2        | 1m03s   | 10x
| Michael Mengel    | Apple II UltraWarp 13Mhz             | Apple II v. 3.2         | Fib2        | 0m21s   | 1x
| Michael Mengel    | Apple II UltraWarp 13Mhz             | Apple GraForth          | Fib2        | 0m11s   | 1x
| Michael Mengel    | Apple II 1Mhz                        | Apple II v. 3.2         | Fib2        | 3m56s   | 1x
| Michael Mengel    | Apple II 1Mhz                        | Apple GraForth          | Fib2        | 2m19s   | 1x
| Wolfgang Stief    | SUN SparcStation 10 TI TMS390255     | OpenFirmware            | Fib2        | 0,2s    |
| Wolfgang Stief    | SUN Ultra 1 200 Mhz UltraSprac       | OpenBoot 3.25           | Fib2        | 0,06s   |
| Stefan Niestegge  | Atari Falcon 68060 100mhz            | f68kans                 | Fib2        | 0,0012s |
| Martin Neitzel    | Asus EeePC 1000h (Atom N270 1.6Ghz)  | FreeBSD 9 FICL Bootloader | Fib2      | 66s     |
| Sabine "Atari Frosch" Engelhardt | Atari Portfolio       | VolksForth 3.81         | Fib2        | 35s     |
| Herbert Lange     | Compaq Deskpro P166                  | pForth V27              | Fib2        | 0,001s  |
| Herbert Lange     | Apple iMac G3 400Mhz                 | pForth V27              | Fib2        | 0,001s  |
| Herbert Lange     | SUN Ultra 1 Creator 3D               | pForth V27              | Fib2        | 0,001s  |
| Ralf Neumann      | mc-CP/M Z80 4Mhz                     | FIG-Forth 1.1           | Fib2        | 1m19s   |
| Ralf Neumann      | Prof80 CP/M Z80 6Mhz                 | FIG-Forth 1.1           | Fib2        | 53s     |
| Carsten Strotmann | IBM L40S (386SX)                     | mina (Fig-Forth)        | Fib2 (1000) | 8s
| Carsten Strotmann | IBM L40X (386SX)                     | F83 (Laxen & Perry)     | Fib2 (1000) | 8s
| Carsten Strotmann | IBM L40X (386SX)                     | GNU Forth 0.5.0 ec8086  | Fib2 (1000) | 24s
| Thorsten Schoeler | Sinclair Spectrum+                   | Aber Forth (FIG-Forth)  | Fib2 (1000) | 1m46s
| Thorsten Schoeler | HX-20                                | Epson Forth E1.0        | Fib2 (1000) | 3m16s
| Thorsten Schoeler | Fignition (ATMEL)                    | Fignition Forth         | fib2 (1000) | 13s
| Thorsten Kuphaldt | C64 (normal)                         | Forth64                 | Fib2 (1000) | 3m50s
| Thorsten Kuphaldt | C64 (Turbo FPGA 6502)                | Forth64                 | Fib2 (1000) | 16s
| Carsten Strotmann | Zilog Super-8 20Mhz                  | Super8 Forth            | Fib2 (1000) | 31s
| Bernd Paysan      | Samsung Galaxy Note 2 (Exynos 4core) | Gforth                  | Fib2 (1000) | 0.01s
| Thorsten Schoeler | PDP11                                | FIG-Forth 1.3           | Fib2 (1000) | 37s
| Norbert Kehrer    | Mupid II (BTX Decoder)               | FIG-Forth 1.1           | Fib2 (1000) | 210s
| Norbert Kehrer    | Mupid II (BTX Decoder)               | Camel Forth 1.01        | Fib 2 (1000)| 150s

[Back to contents](#contents)

### Forth Nesting Benchmark

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/nesting.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/nesting.asm

Use R.A.S. for function:

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/nestingr.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/nestingr.asm

Use data stack for function:

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/nestings.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/nestings.asm

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/nesting.c

|       Name        |              System                  |        Forth / C        |  Benchmark   | Time (sec/round) | Scale |
| :---------------: | :----------------------------------: | :---------------------: | :----------: | :--------------- | :---: |
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH recursive      | Nesting 1mil | 1m2.67s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH                | Nesting 1mil | 34.72s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH use data stack | Nesting 1mil | 16.47s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | zcc z88dk v16209        | Nesting 1mil | 18.78s
| Johan Kotlinski   | C64                                  | DurexForth 1.6.1 (STC)  | Nest 1M      | 17s              | 1x
| Carsten Strotmann | A-ONE (Apple 1 Clone) mit 65C02      | TaliForth 2 (STC)       | Nest 1M      | 25s              | 1x
| Thomas Woinke     | Steckschwein 8MHz 65c02              | TaliForth 2 (STC)       | Next 1M      | 3.7s             | 1x
| Thorsten Schoeler | ZX Spectrum 2+                       | FIG-Forth 1.1a          | Nesting      | 3m15s
| Thorsten Schoeler | ZX Spectrum 2+                       | Spect ZX Forth 1.1      | Nesting      | 4m13s
| Thorsten Schoeler | Sinclair Spectrum+                   | Aber Forth (FIG-Forth)  | Nesting 1m   | 3m17s
| Carsten Strotmann | Zilog Super-8 20Mhz                  | Super8 Forth            | Nesting 1m   | 20s
| Thorsten Schoeler | PDP11                                | FIG-Forth 1.3           | Nesting 1m   | 49s
| Norbert Kehrer    | Mupid II (BTX Decoder)               | FIG-Forth 1.1           | Nesting 1m   | 380s
| Thorsten Schoeler | NCR 3150 486SX/25Mhz+FPU=Linux 2.0.0 | gforth 0.3.0            | Nesting 1m   | 3s
| Norbert Kehrer    | Mupid II (BTX Decoder)               | Camel Forth 1.01        | Nesting 1m   | 292s
| Ingo Soetebier    | Nextstation 68040 33Mhz              | pfe                     | Nesting 1Mil | 340
| KC85 Team         | KC85/4 U880 4Mhz                     | VolksForth CP/M         | Nesting 1Mil | 144
| Thorsten Kuphaldt | Amiga 3000 68030 25Mhz               | jforth                  | Nesting 1Mil | 1.32
| Stefan Herold     | Amstrad 6128+ Z80A 4Mhz              | Uniforth                | Nesting 1Mil | 206
| Ingo Soetebier    | iBook PPC 750lx (G3) 600Mhz          | OpenFirmware            | Nesting 1Mil | 1
| Michael Kalus     | Rockwell R1200-14, 2Mhz 65F12        | RSC-Forth               | Nesting 1Mil | 149
| Matthias Trute    | Atmega16 8MHz                        | amForth 4.4             | Nesting 1Mil | 15.4
| Thorsten Schoeler | HX-20                                | Epson Forth E1.0        | Nesting 1mil | 5m08s
| Martin Neitzel    | Asus EeePC 1000h (Atom N270 1.6Ghz)  |FreeBSD 9 FICL Bootloader| Nesting 1mil | 0.66s
| Helfried Schuerer | robotron K 1510, U 808 D (i8008), 480 kHz (auf FOSY Emulator) | FOSY V1.2P 1988 (FIG) | Nesting 1MILLION | 02:32:05 2h 32min 5s

[Back to contents](#contents)

### Forth Sieve Benchmark

The algorithm finds all odd prime numbers up to limit

    i = 0..8190
    Prime number = 2*i+3

I changed

    : do-prime
      flags size 1+ 1 fill
      0 size 0 do 
        flags I + C@ if
          I dup + 3 + dup I +
          begin 
            dup size 1+ < 
          while 
            0 over flags + C! over + 
          repeat
          drop drop 1+
        then
      loop
      . ." PRIMES" ;

to

    : do-prime
    flags size 1+ 1 fill
    0 flags 3 ( sum_primes i+flags 2*i+3 )
    begin
      over C@ if
        2dup + ( sum_primes i+flags x=2*i+3 3*i+3+flags+n*x )
        begin
          dup flags size + 1+ u< 
          while
            0 over c!
            over +
        repeat
        drop
        rot 1+ -rot
      then
      2 +
      swap 1+ swap
      over flags size + <> while
    repeat
    drop drop
    . ." PRIMES" ;

- http://www.forth.org/TM-10656.pdf

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/sieve.fth
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/sieve.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/sieve.asm

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/sieve_edit.fth
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/sieve_edit.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/sieve_edit.asm

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/sieve_modify.asm

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/sieve.c

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/boriel_sieve.bas

I commented out initializing the field to zero for speed, since the documentation says it should reset itself.

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/boriel_sieve.asm

asm cannot be compiled using pasmo or sjasmplus


|       Name        |                System                |        Forth / C        |  Benchmark   | Time (sec/round) | Scale |
| :---------------: | :----------------------------------: | :---------------------: | :----------: | :--------------- | :---: |
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH                | Sieve Bench  | 0.94s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | M4_FORTH edit           | Sieve Bench  | 0.7s
| Dw0rkin           | ZX Spectrum Fuse 1.6.0 Ubuntu        | modify assembler        | Sieve Bench  | 0.36s
| Dw0rkin           | ZX Spectrum Fuse 1.6.0 Ubuntu        | Boriel Basic zxbc 1.16.4| Sieve Bench  | 3.48s
| Dw0rkin           | ZX Spectrum Fuse 1.5.1 Ubuntu        | zcc z88dk v16209        | Sieve Bench  | 5.94s
| Johan Kotlinski   | C64                                  | DurexForth 1.6.1 (STC)  | Sieve/Prime  | 10s              | 1x
| Alexander Muller  | Raspberry Pi ARM 700Mhz              | Gforth 0.7.0            | Sieve        | 0m08s            | 100x
| Andreas Boehm     | Commodore C64 6510                   | Audiogenic Forth-64     | Sieve Bench  | 18.1             |
| Venty             | Nokia N900 ARM A8 600Mhz             | gforth-fast, Linux      | Sieve Bench  | 0.015            |
| Venty             | Nokia N900 ARM A8 600Mhz             | gforth-dtc, Linux       | Sieve Bench  | 0.025            |
| Venty             | Nokia N900 ARM A8 600Mhz             | gforth-itc, Linux       | Sieve Bench  | 0.028            |
| Thorsten Kuphaldt | Amiga 3000 68030 25Mhz               | jforth                  | Sieve Bench  | 0.148            |
| Stefan Herold     | Amstrad 6128+ Z80A 4Mhz              | Uniforth                | Sieve Bench  | 12               |
| Ingo Soetebier    | iBook PPC 750lx (G3) 600Mhz          | OpenFirmware            | Sieve Bench  | 0.031            |
| Norbert Kehrer    | Mupid II (BTX Decoder)               | FIG-Forth 1.1           | Sieve        | 22s              |
| Norbert Kehrer    | Mupid II (BTX Decoder)               | Camel Forth 1.01        | Sieve        | 15s              |

[Back to contents](#contents)

### FILLIN

Fill the screen with pixels, and set all attributes to 255.

[Performance of Forth](https://spectrumcomputing.co.uk/forums/viewtopic.php?f=6&t=3487 "Spectrum Computing Forums")

In my opinion, using a loop is not the best way to write the solution, because the loop has its own overhead.
We say to make a new variable on the side, and then we copy that into the TOS. So using `BEGIN ... flag UNTIL` I see as exactly what we want.

For best results, I try to limit movement on the stack, such as inserting or removing values.
That is, words that change the total number of values on the stack.
That's why I use words or word combination like `DROP_I`, which overrides TOS, but I also show the worst `DROP I` variant, which moves the stack twice.

Here I copy the results listed in the forum, with an assembler with essentially the same code as my measurement instead of the 0.04 seconds measured by 0.07 seconds.
Plus or minus 0.01 makes an interruption. What causes the rest of the difference I have no idea.
Maybe I'm doing the measurements through POKE in BASIC and that has some small constant overhead.

    Results (Time taken):
    basic : 72.50secs
    Abersoft forth : 1.50secs
    assembly: 0.04secs

| M4 Forth                                                                                                                                                                                                                             | Bytes |  Time  |
| :----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- | :---: |:-----: |
|[PUSH2(23296,16384)<br /> DO<br />&nbsp; &nbsp; PUSH(255)<br />&nbsp; &nbsp; I<br />&nbsp; &nbsp; STORE <br />LOOP                                                                                               ](./fillin_v00.asm)  |   48  | 0.19s  |
|[PUSH2(23296,16384)<br /> DO<br />&nbsp; &nbsp; PUSH(255)<br />&nbsp; &nbsp; I<br />&nbsp; &nbsp; CSTORE <br />LOOP                                                                                              ](./fillin_v01.asm)  |   45  | 0.15s  |
|[PUSH2(23296,16384)<br /> DO<br />&nbsp; &nbsp; PUSH(65535)<br />&nbsp; &nbsp; I<br />&nbsp; &nbsp; STORE <br />PUSH(2) ADDLOOP                                                                                  ](./fillin_v02.asm)  |   48  | 0.12s  |
|[PUSH2(23296,16384)<br /> DO(S)<br />&nbsp; &nbsp; PUSH(65535)<br />&nbsp; &nbsp; I<br />&nbsp; &nbsp; STORE <br />PUSH(2) ADDLOOP                                                                               ](./fillin_v03.asm)  |   53  | 0.07s  |
|[PUSH2(65535,0)<br /> PUSH2(23296,16384)<br /> DO<br />&nbsp; &nbsp; DROP<br />&nbsp; &nbsp; I<br />&nbsp; &nbsp; _2DUP<br />&nbsp; &nbsp; STORE <br />LOOP <br />DROP                                           ](./fillin_v04.asm)  |   58  | 0.22s  |
|[PUSH(65535,23296,16384)<br /> DO(S)<br />&nbsp; &nbsp; _2_PICK<br />&nbsp; &nbsp; OVER<br />&nbsp; &nbsp; STORE <br />LOOP <br />_DROP                                                                          ](./fillin_v05.asm)  |   59  | 0.18s  |
|[PUSH(0,23296,16384)<br /> DO<br />&nbsp; &nbsp; DROP_I<br />&nbsp; &nbsp; PUSH(255) OVER CSTORE <br />LOOP <br />DROP                                                                                           ](./fillin_v06.asm)  |   53  | 0.18s  |
|[PUSH(0,23296,16384)<br /> DO<br />&nbsp; &nbsp; DROP I<br />&nbsp; &nbsp; PUSH(65535) OVER STORE <br />PUSH(2) ADDLOOP<br />DROP                                                                                ](./fillin_v07.asm)  |   58  | 0.14s  |
|[PUSH(65535,0,23296,16384)<br /> DO<br />&nbsp; &nbsp; DROP I<br />&nbsp; &nbsp; _2DUP STORE <br />PUSH(2) ADDLOOP<br />_2DROP                                                                                   ](./fillin_v08.asm)  |   59  | 0.12s  |
|[PUSH(0x4000)<br /> BEGIN<br />&nbsp; &nbsp; PUSH(255) OVER<br />&nbsp; &nbsp; CSTORE _1ADD<br />&nbsp; &nbsp; DUP<br />&nbsp; &nbsp; PUSH(0x5B00)<br />&nbsp; &nbsp; EQ <br />UNTIL <br />DROP                  ](./fillin_v09.asm)  |   42  | 0.12s  |
|[PUSH(0x4000)<br /> BEGIN<br />&nbsp; &nbsp; PUSH(255) OVER CSTORE _1ADD<br />&nbsp; &nbsp; define({_TYP_SINGLE},{L_first})<br />DUP PUSH(0x5B00) EQ UNTIL<br />DROP                                             ](./fillin_v10.asm)  |   46  | 0.09s  |
|[PUSH(0x4000)<br /> BEGIN<br />&nbsp; &nbsp; PUSH(65535) OVER STORE _2ADD<br />DUP PUSH(0x5B00) HEQ UNTIL<br />DROP                                                                                              ](./fillin_v11.asm)  |   44  | 0.07s  |
|[PUSH2(65535,0x4000)<br /> BEGIN<br />&nbsp; &nbsp; _2DUP STORE _2ADD<br />&nbsp; &nbsp; _2DUP STORE _2ADD<br />&nbsp; &nbsp; DUP PUSH(0x5B00) HEQ <br />UNTIL <br />_2DROP                                      ](./fillin_v12.asm)  |   49  | 0.06s  |
|[PUSH2(65535,0x4000)<br /> BEGIN<br />&nbsp; &nbsp; _2DUP_STORE _2ADD<br />DUP_PUSH_EQ_UNTIL(0x5B00) <br />_2DROP                                                                                                ](./fillin_v13.asm)  |   50  | 0.07s  |
|[PUSH2(0x5BFF,0x4000)<br /> BEGIN<br />&nbsp; &nbsp; _2DUP CSTORE _1CADD<br />&nbsp; &nbsp; _2DUP CSTORE _1CADD<br />&nbsp; &nbsp; _2DUP CSTORE _1CADD<br />&nbsp; &nbsp; _2DUP CSTORE _1ADD<br />&nbsp; &nbsp; _2DUP HEQ UNTIL<br />_2DROP](./fillin_v14.asm)  |   48  | 0.06s  |
|[PUSH3(0x4000,6912,255)<br /> FILL                                                                                                                                                                               ](./fillin_v15.asm)  |   47  | 0.06s  |
|[__ASM({<br /> &nbsp; &nbsp; push HL<br /> &nbsp; &nbsp; ld HL, 0xFFFF <br /> &nbsp; &nbsp; ld B, 216<br /> &nbsp; &nbsp; di<br /> &nbsp; &nbsp; ld (\$+7+16+3),SP<br /> &nbsp; &nbsp; ld SP, 0x5B00<br /> &nbsp; rept 16<br /> &nbsp; &nbsp; push HL<br /> &nbsp; endm<br /> &nbsp; &nbsp; djnz $-16<br /> &nbsp; &nbsp; ld SP, 0x0000<br /> &nbsp; &nbsp; ei<br /> &nbsp; &nbsp; pop HL})](./fillin_v16.asm)  |   58  | 0.01s  |

[Back to contents](#contents)

### Project Euler 21: Amicable numbers

- https://projecteuler.net/problem=21
- https://euler.beerbaronbill.com/en/latest/solutions/21.html
- https://users.rust-lang.org/t/executable-size-and-performance-vs-c/4496/3

I came across the last link calculating amicable numbers from the 21st assignment of the Euler project. 
And he used the C source after a little modification. I converted it to a Forth resource.
Using the script `../fth2m4.sh euler_21.fth > euler_21.m4` I converted the Forth source to the M4 FORTH source.

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/euler_21_c.c
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/euler_21_b.bas
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/euler_21.fth
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/euler_21.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/euler_21_fast.m4

After compiling I got these results

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/euler_21_c.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/euler_21_c.c.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/euler_21_b.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/euler_21.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/euler_21_fast.asm


|       Name        |             System             |              Forth / C              |   Benchmark   | Time (sec/round) |   bin size   |
| :---------------: | :----------------------------: | :---------------------------------: | :-----------: | :--------------- | :----------: |
| Dw0rkin           | ZX Spectrum Fuse 1.5.7 Ubuntu  | z88dk.zcc v19766-9ffe2042c-20220722 | euler_21      | 3h 47m 15.41s    | 3314 B (include 4x8 font)
| Dw0rkin           | ZX Spectrum Fuse 1.5.7 Ubuntu  | Boriel Basic zxbc.py 1.16.4         | euler_21      | 5h 0m 42.19s     | 1356 B
| Dw0rkin           | ZX Spectrum Fuse 1.5.7 Ubuntu  | M4_FORTH                            | euler_21      | 1h 47m 12.43s    | 318 B
| Dw0rkin           | ZX Spectrum Fuse 1.5.7 Ubuntu  | M4_FORTH TYPDIV:fast                | euler_21      | 1h 41m 36.85s    | 429 B

[Back to contents](#contents)

### N-body problem

- https://benchmarksgame-team.pages.debian.net/benchmarksgame/description/nbody.html
- https://github.com/z88dk/z88dk/wiki/Benchmarks#n-body
- https://github.com/z88dk/z88dk/blob/master/libsrc/_DEVELOPMENT/EXAMPLES/benchmarks/n-body/n-body.c

This is a floating point arithmetic test.
I made a small optimization of the "advance" procedure, where I combined two loops into one.
I rewrote the C source to forth. 
I used the fth2m4.sh script to convert to M4 FORTH.
I made 2 versions, one with 48 bit floating point numbers using the ZX Spectrum ROM (separate stack for real numbers) and the other a fast 16 bit floating point number in DANagy format.

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/n-body_c.c
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/n-body_f.fth
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/n-body_z.fth (separate stack for real numbers)

      ../fth2m4.sh         n-body_f.fth > n-body_f.m4
      ../fth2m4.sh -zfloat n-body_z.fth > n-body_z.m4

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/n-body_f.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/n-body_z.m4

After compiling...

    z88dk.zcc +zx -lndos -O2      -lm -create-app n-body_c.c -o n-body_c_48b.bin 
    z88dk.zcc +zx -lndos -O2 --math32 -create-app n-body_c.c -o n-body_c_32b.bin

    z80dasm -g 32768 -a n-body_c_48b.bin > n-body_c_48b.asm
    z80dasm -g 32768 -a n-body_c_16b.bin > n-body_c_16b.asm
    
    ../compile.sh n-body_f 32768
    ../compile.sh n-body_z 32768

I got these results:
 
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/n-body_c_48b.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/n-body_c_32b.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/n-body_f.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/n-body_z.asm


|   Name    |                      System                     |              Forth / C              |    Parameters     | Benchmark | mantissa | Time (round)  |   bin size   |  Scale   |
| :-------: | :---------------------------------------------: | :---------------------------------: | :---------------: | :-------: | :------: | :-----------: | :----------: | :------: |
| Dw0rkin   |          ZX Spectrum Fuse 1.5.7 Ubuntu          | z88dk.zcc v19766-9ffe2042c-20220722 | -O2 -lm           | n-body_c  | 40-bit   | 17m 43.67s    | 5872 B       | 1
| Dw0rkin   |          ZX Spectrum Fuse 1.5.7 Ubuntu          | z88dk.zcc v19766-9ffe2042c-20220722 | -O2 --math16      | n-body_c  | 10-bit   | compiled fail | 0 B          | 1
| Dw0rkin   |          ZX Spectrum Fuse 1.5.7 Ubuntu          | z88dk.zcc v19766-9ffe2042c-20220722 | -O2 --math32      | n-body_c  | 22-bit   |  4m 55.24s    | 7319 B       | 1
| Dw0rkin   |          ZX Spectrum Fuse 1.5.7 Ubuntu          | M4_FORTH                            |                   | n-body_z  | 31-bit   | 26m  6.85s    | 1891 B       | 1
| Dw0rkin   |          ZX Spectrum Fuse 1.5.7 Ubuntu          | M4_FORTH                            |                   | n-body_f  |  8-bit   |  0m 23.16s    | 4174 B       | 1
| Dw0rkin   | AMD Ryzen 9 4900H Ubuntu 11.4.0-1ubuntu1~22.04  | gcc 11.4.0                          |                   | n-body_c  | 52-bit   |  0m 10.41s    | 16512 B      | 50000
| Dw0rkin   | AMD Ryzen 9 4900H Ubuntu 11.4.0-1ubuntu1~22.04  | gcc 11.4.0 -O2                      | -O2               | n-body_c  | 52-bit   |  0m  4.42s    | 16520 B      | 50000
| Dw0rkin   | AMD Ryzen 9 4900H Ubuntu 11.4.0-1ubuntu1~22.04  | gforth 0.7.3                        |                   | n-body_z  | 52-bit   |  1m 23.98s    | interpreter  | 50000


[Back to contents](#contents)

### Pangram Benchmark

    : pangram? ( addr len -- ? )
      0 -rot bounds do
        i c@ 32 or [char] a -
        dup 0 26 within if
          1 swap lshift or
        else drop then
      loop
      1 26 lshift 1- = ;

    s" The five boxing wizards jump quickly." pangram? .   \ -1

10000x + 1x with print output

[Back to contents](#contents)

#### Forth from Rosettacode

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/Pangram_rosettacode.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/Pangram_rosettacode.asm

But this code from rosettacode has one trick. The cell is assumed to be 32 bits.
When rewriting into double cells, we get into a non-standard area. There is no word for double OR - DOR. There is no word for a double logical left shift - DLSHIFT.

    COLON(_pangram_,( addr len -- ? ))
      PUSHDOT_2SWAP(0) OVER_ADD SWAP DO
        I CFETCH PUSH_OR(32) PUSH_SUB('a')
        DUP_PUSH2_WITHIN_IF(0,26)
          PUSHDOT(1) ROT_DLSHIFT DOR
        ELSE DROP THEN
      LOOP
      PUSHDOT_DEQ(0x3FFFFFF) SEMICOLON

[Back to contents](#contents)

#### Double BITSET version

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/Pangram_dbitset.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/Pangram_dbitset.asm

Since `1. ROT DLSHIFT DOR` is actually just a bit setting, I wrote an optimized `DSETBIT` word for it.

And replaced the DO LOOP loop with the more efficient `BEGIN WHILE REPEAT`.

That crazy combination of words just says that it should increase the pointer by one and load the character, if it is zero then jump out of the loop, otherwise save a copy of the character in TOS.

The `ROT` moves the third cell to the TOS. So it's a `SWAP` between a 16-bit and a 32-bit number.

`NROT` moves TOS to third position. So it's a `SWAP` between a 16-bit and a 32-bit number.

`ROT ... NROT` temporarily moves NNOS (third cell) to TOS. Similar to repeated `SWAP`, but between 16-bit and 32-bit.

`_2OVER_NIP` creates a copy of NNOS (third cell) to TOS. Similar to `OVER`, but between 16-bit and 32-bit. It seems that there are no standard word for this.

The less the data stack moves up and down between words, the faster and more efficient the program is.
Most compound words use this.
Like for example `SWAP_1ADD_SWAP` does not need to move the stack internally.
It's easy to check this in the console using the check_word.sh script.

    COLON(_pangram_,( addr -- ? ))
      _1SUB
      PUSHDOT(0)
      BEGIN
        ROT_1ADD_NROT_2OVER_NIP_CFETCH_0CNE_WHILE_2OVER_NIP_CFETCH
        PUSH_OR(32) PUSH_SUB('a')
        DUP_PUSH2_WITHIN_IF(0,26)
          DBITSET
    dnl #  _2OVER_NIP_CFETCH EMIT SPACE_2DUP_HEX_UDDOT CR
        ELSE DROP THEN
      REPEAT
      PUSHDOT_DEQ(0x3FFFFFF) NIP SEMICOLON

[Back to contents](#contents)

#### Double BITSET inline version

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/Pangram_dbitset_inline.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/Pangram_dbitset_inline.asm

I only defined at the beginning define({_TYP_DOUBLE},{fast})

In this case, it changed the DBITSET function to inline, so it shortened it because it was only called from one place.

[Back to contents](#contents)

#### C version

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/Pangram.c
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/Pangram.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/Pangram.lst

This is really scary if you look at what kind of code it is. That would be worth writing a whole article about.
It's hard just figuring out how to compile it and I had to help with the asm function to print the number anyway.
Because printf("%i\n") reboots the ZX Spectrum.

It took me a while to figure out how to write it without mistakes. Because it requires a different notation than pasmo.
It requires a destination registry, where pasmo only wants the source registry, because there is no other option to confuse it with.
Maybe it's a consequence that it's not only targeted at the Z80, but also at other processors or more modern variants that can also do other instructions.

    xor   a, a          ; 1:4       prt_s16   neg
    sub   a, l          ; 1:4       prt_s16   neg

It didn't take decimal numbers or 'a' characters, just this format and it reported a completely different error.

    ld   bc, #0xD8F0    ; 3:10      prt_u16   -10000
    or    a, #0x30      ; 2:7       bin16_dec   1..9 --> '1'..'9', unchanged '0'..'9'

But he doesn't mind this.

    rst   0x10          ; 1:11      bin16_dec   putchar with ZX 48K ROM in, this will print char in A

Labels must only have this numeric format.

    jr   nc, 00001$     ; 2:7/12    prt_s16



|       Name        |              System              |         Forth / C         |           Benchmark         | Time (sec/round) | Scale |
| :---------------: | :------------------------------: | :-----------------------: | :-------------------------: | :--------------- | :---: |
| Dw0rkin           | ZX Spectrum Fuse 1.5.7 Ubuntu    | M4_FORTH                  | Pangram do loop             | 50.41s
| Dw0rkin           | ZX Spectrum Fuse 1.5.7 Ubuntu    | M4_FORTH                  | Pangram begin bitset repeat | 32.26s
| Dw0rkin           | ZX Spectrum Fuse 1.5.7 Ubuntu    | M4_FORTH _TYP_DOUBLE:fast | Pangram begin bitset repeat | 27.58s
| Dw0rkin           | ZX Spectrum Fuse 1.5.7 Ubuntu    | sdcc 3.8.0 #10562 (Linux) | Pangram                     | 2m 51.61s

[Back to contents](#contents)


### Logical conditions with 8, 16 and 32 bit constants

This was not originally supposed to be a benchmark, but a test of the correctness of the execution of logical conditions with two variables, where one of them is a constant known at the time of compilation.
I test the conditions <,<=,>,>= either unsigned (the test starts with the letter "u") or with a sign (the test starts with the letter "s").
8-bit, 16-bit and 32-bit constants are tested.
Tests store all results in auxiliary variables in memory and are performed in such a way that in the end their sum is equal to the number of passes through the loop, i.e. overflow to zero.
The constant is compared to the index of the loop, which runs over all values except the 32-bit variant.
A 32-bit value is generated from a 16-bit loop to avoid waiting a year for the result.
0xABCD is converted to 0xABBBBBCD.

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/u8cond_c.c
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/s8cond_c.c
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/u16cond_c.c
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/s16cond_c.c
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/u32cond_c.c
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/s32cond_c.c
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/u8cond.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/s8cond.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/u16cond.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/s16cond.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/u32cond.m4
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/s32cond.m4

After compiling I got these results

- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/u8cond_c.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/s8cond_c.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/u16cond_c.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/s16cond_c.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/u32cond_c.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/s32cond_c.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/u8cond.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/s8cond.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/u16cond.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/s16cond.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/u32cond.asm
- https://codeberg.org/DW0RKiN/M4_FORTH/src/branch/master/Benchmark/s32cond.asm


|    Name    |             System            |               Forth / C              |  Benchmark  |  Time       |           bin size          |
| :--------: | :---------------------------: | :----------------------------------: | :---------: | :---------- | :-------------------------: |
| Dw0rkin    | ZX Spectrum Fuse 1.6.0 Ubuntu | z88dk.zcc v22110-51889e5300-20231219 | -O2 u8cond  |  0m  1.80s  | 11 615 B (include 4x8 font)
| Dw0rkin    | ZX Spectrum Fuse 1.6.0 Ubuntu | z88dk.zcc v22110-51889e5300-20231219 | -O2 s8cond  |  0m  3.42s  | 11 562 B (include 4x8 font)
| Dw0rkin    | ZX Spectrum Fuse 1.6.0 Ubuntu | z88dk.zcc v22110-51889e5300-20231219 | -O2 u16cond |  9m 55.88s  | 12 311 B (include 4x8 font)
| Dw0rkin    | ZX Spectrum Fuse 1.6.0 Ubuntu | z88dk.zcc v22110-51889e5300-20231219 | -O2 s16cond | 12m  1.65s  | 12 249 B (include 4x8 font)
| Dw0rkin    | ZX Spectrum Fuse 1.6.0 Ubuntu | z88dk.zcc v22110-51889e5300-20231219 | -O2 u32cond | 24m 15.85s  | 12 609 B (include 4x8 font)
| Dw0rkin    | ZX Spectrum Fuse 1.6.0 Ubuntu | z88dk.zcc v22110-51889e5300-20231219 | -O2 s32cond | result fail | 12 575 B (include 4x8 font)
| Dw0rkin    | ZX Spectrum Fuse 1.6.0 Ubuntu | M4_FORTH                             | u8cond      |  0m  0.62s  |  4 823 B
| Dw0rkin    | ZX Spectrum Fuse 1.6.0 Ubuntu | M4_FORTH                             | s8cond      |  0m  0.68s  |  5 005 B
| Dw0rkin    | ZX Spectrum Fuse 1.6.0 Ubuntu | M4_FORTH                             | u16cond     |  3m 31.45s  |  5 456 B
| Dw0rkin    | ZX Spectrum Fuse 1.6.0 Ubuntu | M4_FORTH                             | s16cond     |  3m 55.76s  |  5 836 B
| Dw0rkin    | ZX Spectrum Fuse 1.6.0 Ubuntu | M4_FORTH                             | u32cond     |  5m 19.84s  |  6 327 B
| Dw0rkin    | ZX Spectrum Fuse 1.6.0 Ubuntu | M4_FORTH                             | s32cond     |  5m 19.74s  |  7 361 B

[Back to contents](#contents)
