[DEFINED] M4_FORTH [IF]
  define({USE_FONT_5x8})
[ELSE]
  51 constant max_x
  variable x 0 x !
  : cr 0 x ! cr ;
  : plus_x ( -- )
    x @ 1+ dup x ! 
    max_x >= if
      cr
    then
  ;
  : white   .\" \e[0;37m" ;
  : l_blue  .\" \e[1;34m" ;
  : red     .\" \e[0;31m" ;
  : purple  .\" \e[0;35m" ;
  : green   .\" \e[0;32m" ;
  : cyan    .\" \e[0;36m" ;
  : l_brown .\" \e[1;33m" ;
  : black   .\" \e[0;30m" ;
  : other   .\" \e[m" ;
  create cctable
  ' white ,
  ' l_blue ,
  ' red ,
  ' purple ,
  ' green ,
  ' cyan ,
  ' l_brown ,
  ' black ,
  ' other ,
  variable case_emit
  variable default_emit
  : reset_emit default_emit @ case_emit ! ;
  : emit_color ( idx )  255 and 0 max 8 min cells cctable + @ execute reset_emit ;
  : emit_char ( char -- )
    dup emit 
    255 and 
    dup 32 >= if 
      plus_x
    else dup 10 = if
      0 x !    
    else dup 16 = if
      ['] emit_color case_emit !
    then then then
    drop
  ;
  ' emit_char default_emit !
  reset_emit
  : emit ( char -- )
    case_emit @ execute 
  ;
  : type
    dup 0= if 2drop exit then
    1-
    for 
      dup @ emit 1+
    next 
    drop 
  ;
  : .\" POSTPONE s\" POSTPONE TYPE ; IMMEDIATE
  require random.fs
  : -! dup @ rot - swap ! ;
  : u/ swap abs swap abs / ;
  : setseed 255 and for rnd drop next ;
  cr cr cr
[ENDIF]

( initialize variables)
  variable starved            0    starved !
  variable immigrants         5    immigrants !
  variable population         100  population !
  variable harvest            3000 harvest ! ( bushels)
  variable bushels_per_acre   3    bushels_per_acre ! ( harvest yield)
  variable rats_ate           200  rats_ate ! ( bushels)
  variable storage            2800 storage ! ( bushels)
  variable land               1000 land ! ( acres)
  variable cost_per_acre      19   cost_per_acre ! ( bushels per acre)
  variable plague_deaths      0    plague_deaths !
  
  variable acres_planted
  variable bushels_fed
  
: input ( a_message len_message -- u )
    type cr
    begin
      s\"           " ( -- astr ulen )
      over swap       ( astr ulen -- astr astr ulen )
      accept          ( astr astr ulen -- astr uload )
      s>number?       ( astr uload -- d bool )
      .\"   \20\0\n"
    0= while          ( d bool -- d )
        2drop         ( d -- )
        .\" \20\3[Please re-enter a valid number.]\n"
    repeat
    d>s               ( d -- s )
;

[DEFINED] M4_FORTH [IF]
[ELSE]

: ask_LAND
(    123456789abcdef0123456789abcdef0123456789abcdef0123)
  begin
    dup 0 < if
      negate 
      land @ 2dup u> if
( ? [land])
        .\"  * O great Khammurapi,\n"
        .\"    we have but "
        u. drop .\" acres of land!\n"
        s\" Again, how much land will you \20\1SELL?" input
      else
        over - land !
        cost_per_acre @ * storage +!
        exit
      then
    else
      dup cost_per_acre @ * storage @ 2dup u> if
( ? ?*[cost_per_acre] [storage])
        .\"  * O great Khammurapi,\n"
        .\"    we have but "
        dup u. 
        .\" bushels of grain!\n"
        .\"    The maximum is " cost_per_acre @ u/ u. 2drop .\" acres.\n"
        s\" Again, how much land will you \20\1BUY?" input
      else
        swap - storage !
        land +!
        exit
      then 
    then
  again
;

: ask_PLANT
(    123456789abcdef0123456789abcdef0123456789abcdef0123)
  begin
    land @ 2dup u> if
( ? [land] )
      .\"  * O great Khammurapi,\n"
      .\"    we have but "
      u. drop .\" acres of land!\n"
      s\" Again, how much land will you \20\4PLANT?" input
    else drop population @ 10 * 2dup u> if
( ? [population]*10 )
      .\"  * O great Khammurapi,\n"
      .\"    we have but " 
      population @ u. .\" people that can farm the land!\n"
      .\"    The maximum is " u. drop .\" acres.\n"
      s\" Again, how much land will you \20\4PLANT?" input
    else drop storage @ 2 / 2dup u> if
( ? [storage]/2 )
      .\"  * O great Khammurapi,\n"
      .\"    we have but "
      storage @ u. .\" bushels for planting!\n"
      .\"    The maximum is " u. drop .\" acres.\n"
      s\" Again, how much land will you \20\4PLANT?" input
    else
      drop 
      exit
    then then then
  again
;
[ENDIF]

: game
( read timer)
  utime d>s
( ===================== Introductory message ===================== )
(     123456789abcdef0123456789abcdef0123456789abcdef0123___456789abcdef0123456789abcdef0123456789abcdef0123___456789abcdef0123456789abcdef0123456789abcdef0123___456789abcdef0123456789abcdef0123456789abcdef0123___456789abcdef0123456789abcdef0123456789abcdef0123___456789abcdef0123456789abcdef0123456789abcdef0123___456789abcdef0123456789abcdef0123456789abcdef0123)
  .\" \x22Congratulations, you are the new ruler of the an- cient Sumerian city-state of Babilim (The Gate of  Gods), elected for a 12-year term. Your duties are to dispense food, direct farming, buy or sell land as needed to support your people. Watch out for ratinfestations and the plague! Grain is the general  currency, measured in bushels. Basic information:\n\n"

  .\"  * Each person needs at least 20 bushels of grain     per year to survive.\n\n"
  .\"  * Each person can farm at most 10 acres of land.\n\n"
  .\"  * It takes 2 bushels of grain to farm an acre of     land.\n\n"
  .\"  * The market price for land fluctuates yearly.\n\n"
  .\" Rule wisely and you will be showered with apprecia-tion at the end of your term. Rule poorly and you  will be kicked out of office!\x22\n\n"
( wait for scroll and read timer again)
  utime d>s or setseed

( ===================== Annual loop ===================== )
( begin a twelve year loop and print the outcome of each year)
  13 1 do 
( --------------------- Annual summary ---------------------)
(    123456789abcdef0123456789abcdef0123456789abcdef0123
                                                        123456789abcdef0123456789abcdef0123456789abcdef0123)
    .\" O great Khammurapi!\n"
    .\" You are in year " i u. .\" of your twelve-year rule.\n"
    starved @ dup if
      .\" In the previous year " u. .\" people starved to death.\n"
    else
      drop
    then
    immigrants @ dup if
      .\" In the previous year " u. .\" people immigrated.\n"
    else
      drop
    then    
    .\" Now the \20\2population is " population @ u. .\" \20\0.\n"
    .\" We harvested " harvest @ u. .\" bushels at " bushels_per_acre @ u. .\" bushels per acre.\n"   
    rats_ate @ dup if
      .\" Rats destroyed " u. .\" bushels, leaving  \20\1" storage @ u. .\" bushels\n\20\0in storage.\n"
    else
      drop
      .\" Stores contain \20\1" storage @ u. .\" bushels \20\0of grain.\n"
    then
    .\" The city owns \20\4" land @ u. .\" acres \20\0of land.\n"
    .\" Land is currently worth " cost_per_acre @ u. .\" bushels per acre.\n"
    plague_deaths @ dup if
      .\" There were " u. .\" deaths from the plague.\n"
    else
      drop
    then
    cr


( ===================== Questions ===================== )

( --------------------- How many LAND ---------------------)
  s\" How many acres of LAND will you \20\1+BUY or -SELL?" input
  [DEFINED] M4_FORTH [IF]
(    123456789abcdef0123456789abcdef0123456789abcdef0123)
  begin
    dup 0 < if
      negate 
      land @ 2dup u> if
( ? [land])
        .\"  * O great Khammurapi,\n"
        .\"    we have but " 
        u. drop s\" acres of land!\n"
        s\" Again, how much land will you \20\1SELL?" input
      else
        over - land !
        cost_per_acre @ * storage +!
        break
      then
    else
      dup cost_per_acre @ * storage @ 2dup u> if
( ? ?*[cost_per_acre] [storage])
        .\"  * O great Khammurapi,\n"
        .\"    we have but "
        dup u. 
        .\" bushels of grain!\n"
        .\"    The maximum is " cost_per_acre @ u/ u. 2drop .\" acres.\n"
        s\" Again, how much land will you \20\1BUY?" input
      else
        swap - storage !
        land +!
        break
      then 
    then
  again
  [ELSE]
  ask_LAND
  [ENDIF]


( --------------------- How many FEED ---------------------)
  storage @
  s\" How many bushels do you wish to \20\2FEED your people?" input
  begin
( [storage] ?)
    2dup u< while
(    123456789abcdef0123456789abcdef0123456789abcdef0123)
      .\"  * O great Khammurapi,\n"
      .\"    we have but "
      u. .\" bushels of grain!\n"
      .\"    At least " 
      population @ 20 * u. 
      .\" bushels will prevent Famine.\n"
      s\" Again, how much grain will you \20\2FEED the people?" input
  repeat
  dup bushels_fed !
  - storage !


( --------------------- How many PLANT ---------------------)
  s\" How many acres do you wish to \20\4PLANT with seed?" input
  [DEFINED] M4_FORTH [IF]
(    123456789abcdef0123456789abcdef0123456789abcdef0123)
  begin
    land @ 2dup u> if
( ? [land] )
      .\"  * O great Khammurapi,\n"
      .\"    we have but "
      u. drop .\" acres of land!\n"
      s\" Again, how much land will you \20\4PLANT?" input
    else drop population @ 10 * 2dup u> if
( ? [population]*10 )
      .\"  * O great Khammurapi,\n"
      .\"    we have but "
      population @ u. .\" people that can farm the land!\n"
      .\"    The maximum is " u. drop .\" acres.\n"
      s\" Again, how much land will you \20\4PLANT?" input
    else drop storage @ 2 / 2dup u> if
( ? [storage]/2 )
      .\"  * O great Khammurapi,\n"
      .\"    we have but " 
      storage @ u. .\" bushels for planting!\n"
      .\"    The maximum is " u. drop .\" acres.\n"
      s\" Again, how much land will you \20\4PLANT?" input
    else
      drop 
      break
    then then then
  again
  [ELSE]
  ask_PLANT
  [ENDIF]
  dup acres_planted !
  2 * storage -!


( ===================== Consequences =====================)

( --------------------- check PLAGUE ---------------------)
  20 random ( 0..19)
  3 u< if
    population @ 2 u/
(      123456789abcdef0123456789abcdef0123456789abcdef0123)
    s\"  * O great Khammurapi,\n" type
    s\"    unluckily a horrible PLAGUE happened!\n" type 
  else
    0
    s\"  * Thank God! No plague this year\n" type
  then
  dup plague_deaths !
  population -!


( --------------------- check STARVED [ -- starved ] ---------------------)
  population @ dup bushels_fed @ 20 u/ -
  dup 0 > if
( [population] +starved)
(  half of those starving will die)
    2 u/
(      123456789abcdef0123456789abcdef0123456789abcdef0123)
    s\"  * O great Khammurapi,\n" type
    s\"    unfortunately we have " type dup u. s\" PEOPLE STARVED\n" type
    2dup
( [population] starved [population] starved  )
( checking if more than 45 percent of the population died)
    100 * swap 45 * 
( [population] starved  100*starved 45*[population] )
    u> if
(        123456789abcdef0123456789abcdef0123456789abcdef0123)
      s\"  * O great Khammurapi,\n" type
      s\"    we really regret to inform you that this year\n   " type
(                                                            ***456789abcdef0  123456789abcdef0123456789abcdef0123***456789abcdef0123456789abcdef0123456789abcdef0123***456789abcdef0123456789abcdef0123456789abcdef0123)
      u. drop s\" people have been starved, which represent\n   more than 45% of total population in your city-    state. Unfortunately, according to our laws, you   were immediately deposed.\n" type
(        123456789abcdef0123456789abcdef0123456789abcdef0123)
      s\" ============== G A M E ===== O V E R ==============" type
      s\"                 O great Khammurapi,\n" type
      s\"       we're truly regretful that in year " type i u. s\" of\n" type
      s\"               your twelve year rule,\n" type
      s\"          you lost your imperial authority.\n\n" type
      s\"      *** THANK YOU FOR PLAYING KHAMMURAPI! ***\n" type
      bye
    else
      tuck - population !
    then
  else
    2drop 0
    s\"  * O great Khammurapi,\n" type
    s\"    This year we have NO people starved\n" type
  then
  dup starved !
  

    
( --------------------- check IMMIGRANTS [ starved -- ] ---------------------)
(       123456789abcdef0123456789abcdef0123456789abcdef0123***456789abcdef0123456789abcdef0123456789abcdef0123)
  if
    0
    s\"  * O great Khammurapi,\n" type
    s\"    due to starvation, NO IMMIGRANTS this year come    to our city-state Babilim\n" type
  else
( {20 * land + storage} / {100 * population} + 1)
    storage @ land @ 20 * + population @ 100 * u/ 1+
(       123456789abcdef0123456789abcdef0123456789abcdef0123***456789abcdef0123456789abcdef0123456789abcdef0123)
    s\"  * O great Khammurapi,\n" type
    s\"    this year " type
    dup u. 
                     s\" immigrants come to our Babilim\n" type
  then
  dup immigrants !        
  population +!

  
( --------------------- check RAT_ATE before harvest! ---------------------)
( 40 percent a rat infestation happened like original)
  100 random 40 < if 
    storage @ 30 random 2 +
    s\"  * O great Khammurapi,\n" type
(      123456789abcdef0123456789abcdef0123456789abcdef0123   *                                               123456789abcdef0123456789abcdef0123456789abcdef0123
          *                                               123456789abcdef0123456789abcdef0123456789abcdef0123   *)
    dup 2 = if
    s\"    unfathomably, a nightmarish RAT INFESTATION has    struck, decimating half of the stored supplies     with merciless precision.\n" type
    else dup 3 = if
    s\"    appallingly, a horrifying RAT INFESTATION un-      folded, resulting in the severe destruction of a   third of the stockpiled provisions.\n" type
    else dup 4 = if
    s\"    tragically, a devastating RAT INFESTATION occur-   red, leaving in its wake the substantial ruin of   a quarter of the stored supplies.\n" type
    else dup 10 u< if
    s\"    regrettably, an alarming RAT INFESTATION has ta-   ken place, causing the significant destruction,    specifically up to 20%, of the stored resources." type
    else dup 20 u< if
    s\"    unfortunately, a concerning RAT INFESTATION hap-   pened, with the noteworthy consequence of up to    10% of the stockpiled supplies being damaged.\n" type
    else
    s\"    unluckily a RAT INFESTATION happened and the de-   struction of up to 5% of the stored stock.\n" type
    then then then then then 
( [storage] 2..31 )
    u/
( [storage]/2 .. [storage]/31 rats_ate from storage --> an average of 10 percent)       
  else
    0
    s\"  * Thank God! No rat infestation occurred this year" type
  then
  dup rats_ate !
  storage -!


( --------------------- check BUSHELS_PER_ACRE [ -- bushels_per_acre] ---------------------)
  5 random 5 random + 
( 0..8, original is 1..5)
  dup 0 = if
(       123456789abcdef0123456789abcdef0123456789abcdef0123)
    s\"  * O great Khammurapi,\n" type
    s\"    terrible calamity, LOCUSTS devoured the crops.\n" type   
  else dup 3 < if
    s\"  * O great Khammurapi,\n" type
    s\"    we were hit by a major crop failure.\n" type
  else dup 6 > if
    s\"  * O brilliant Khammurapi,\n" type
    s\"    we have been blessed with a great harvest.\n" type
  then then then
  dup bushels_per_acre !
  

( --------------------- check HARVEST [ bushels_per_acre -- ] ---------------------)
  acres_planted @ * 
  dup harvest !
  dup storage +!
(    123456789abcdef0123456789abcdef0123456789abcdef0123)
  s\"  * O great Khammurapi,\n" type
  s\"    this year we harvest " type u. s\" bushels of grain\n" type
  
( --------------------- check COST_PER_ACRE ---------------------)
( 17..26 like original)
  10 random 17 + 
(    123456789abcdef0123456789abcdef0123456789abcdef0123)
  s\"  * O great Khammurapi,\n" type
  s\"    the price of land will be " type dup u. s\" bushels per acre\n   next year\n" type
  cost_per_acre !


( ---- debug) 
depth dup if
  .\" \nOh no... stack contains " . .\" items!\n"
then
drop
                      

( ===================== End of year loop =====================)
  loop


( ===================== Final summary =====================)

( --------------------- Sumarize result ---------------------)
(     123456789abcdef0123456789abcdef0123456789abcdef0123)
  .\" ========= C O N G R A T U L A T I O N S ! ========="
  .\"             Khammurapi your highness,\n"
  .\"    you've successfully gone through your twelve\n"
  .\"                year term of office.\n"
  .\"          In the end of your term of reign,\n"
  .\"                the Sumer Kingdom:\n\n"
  .\"   * owns "
  land @ dup u. .\" acres of wealthy land;\n"
  .\"   * owns " 
  storage @ dup u. .\" bushels of grain in storage;\n"
  .\"   * has a total population of "
  population @ dup u.  .\" ;\n"
  .\"   * has " 
  starved @ dup u. .\" people starved.\n\n"
( land storage population starved )


( --------------------- Last starved ---------------------)
(       123456789abcdef0123456789abcdef0123456789abcdef0123___456789abcdef0123456789abcdef0123456789abcdef0123)
  dup if
    .\" In the end of your term"  
    dup 70 u> if
                           s\"  more than 70 citizens star-ved to death. That's an extremely serious problem!\n"
    else dup 30 u> if
                           s\"  at least 30 people died    from starvation. Every citizen's basic living re-  quirements should be met!\n"
    else
                           s\" , still some people starved to die. This could have been avoided by a reasona- ble farming policy.\n"
    then then
  else
     s\" That's great because now no one starved in our     country!\n"
  then
  type
  drop
( land storage population )


( --------------------- Final population ---------------------)
(       123456789abcdef0123456789abcdef0123456789abcdef0123___456789abcdef0123456789abcdef0123456789abcdef0123)
  dup 300 u> if
    s\" Under your government, the population has more thantripled. Brilliant work.\n"
  else dup 200 u> if
    s\" There are more than twice people now living in     Sumer as twelve years ago. It is well.\n"
  else dup 150 u> if
    s\" During your term of office, population here in-    creases pretty steadily.\n"
  else dup 100 u>= if
    s\" During the twelve years, total number of people in Sumer changed little.\n"
  else dup 50 u>= if
    s\" At the end of your reign, the population is lower  than at the beginning. That's not good.\n"
  else
    s\" Not even half of the population survived your re-  ign... Tragic.\n"
  then then then then then
  type
  drop
( land storage )



( --------------------- Final bushels in storage ---------------------)
(       123456789abcdef0123456789abcdef0123456789abcdef0123___456789abcdef0123456789abcdef0123456789abcdef0123)
  dup 11200 u> if
    s\" Nice job! You have earned a lot to feed and financeour homeland. That's really cool!\n"
  else dup 5600 u> if
    s\" Steady increase of bushels. Our country is wealthy!"
  else dup 2800 u> if
    s\" Fine policies, but haven't earned much. We need    more bushels to thrive!\n"
  else
    s\" Oops, you haven't handled your financial task well.So much grain lost!\n"
  then then then
  type
  drop
( land )


( --------------------- Final acres owned ---------------------)
(     123456789abcdef0123456789abcdef0123456789abcdef0123___456789abcdef0123456789abcdef0123456789abcdef0123)
  .\" O great Khammurapi, " 
  dup 1500 u> if
                      s\" you have maintained the growth of the Babilim very well. Fantastic!\n"
  else dup 1000 u> if
                      s\" you have kept our land well for12 years, though our domain has expanded little!\n"
  else dup 500 u> if
                      s\" our fruitful land has shrunk   under your rule. You could have done better!\n"
  else
                      s\" you have lost more than half ofour land! That is a shame. This is really not good!"
  then then then
  type
  drop

  
(     123456789abcdef0123456789abcdef0123456789abcdef0123)
  s\" ====== G A M E ===== A C C O M P L I S H E D ======\n" type
  s\"      *** THANK YOU FOR PLAYING KHAMMURAPI! ***\n" type
;

game
