## Jak vylepšit překladač pro vlastní potřeby?

- [eng version](https://codeberg-org.translate.goog/DW0RKiN/M4_FORTH/src/branch/master/HOWTO.md?_x_tr_sl=cs&_x_tr_tl=en)
- back to the Czech version, use the button to return to the previous page

## Obsah

- [Obsah](#obsah)
- [1. Základy makroprocesoru M4](#1-základy-makroprocesoru-m4)
- [2. Deaktivace maker obalením](#2-deaktivace-maker-obalením)
- [3. Nebezpečné znaky v parametrech makra](#3-nebezpečné-znaky-v-parametrech-makra)
- [4. Odsazení v M4](#4-odsazení-v-m4)
- [5. Definice maker v M4](#5-definice-maker-v-m4)
- [6. Tokenizace a pole __TOKEN[]](#6-tokenizace-a-pole-__token)
- [7. Pomocná makra pro práci s polem __TOKEN[]](#7-pomocná-makra-pro-práci-s-polem-__token) 
- [8. Tokenová pravidla](#8-tokenová-pravidla) 
- [9. Předávané info v makrech](#9-předávané-info-v-makrech) 
- [10. Debugování](#10-debugování)

### 1. Základy makroprocesoru M4

M4 prochází postupně text a hledá jména maker, které odpovídají C syntaxi pro jména proměnných a funkcí.
Začínají znakem `a-z` nebo `A-Z` bez dikaritiky, popřípadě podržítkem `_` a pokračují tím samým ale navíc mohou obsahovat znak `0-9`.
Makra mohou mít parametry předávané v závorce a oddělené čárkami.
Jakýkoliv jiný znak jako je konec řádku, mezera a nebo znak `á` je chápán jako oddělovač jmen maker a zůstane na výstupu.

Definice nového makra `SWAP` vypadá takto:

    define(`SWAP', `
        ex   DE, HL         ; 1:4       swap   ( b a -- a b )')

Odteď je makro aktivní a následující `SWAP` nebo `SWAP()` se nahradí za `ex DE, HL`.
`SWAPSWAP` ale nenajde, muselo by se to oddělit například `SWAP SWAP`. 
Pokud nechceme přidávat mezeru na výstup, tak lze použít závorky pro zadání prázdného parametru 

    SWAP()SWAP

Popřípadě 

    SWAP(parametr je stejne ignorován)SWAP

Poslední možnost je strčít mezi jména prázdný řetezec 

    SWAP`'SWAP

[Zpět na obsah](#obsah)

### 2. Deaktivace maker obalením

Pokud potřebujeme vypsat slovo, které je zároveň jménem makra musíme text obalit do uvozovek. 
Standartní M4 používá tyto:

    ` ...je uvozující znak pro počátek řetezce (najdete ho asi s pomocí pravého altu vlevo od klavesy 1)
    ' ...je ukončující znak pro konec řetezce (apostrof)

Pokud M4 narazí na takto obalený text tak odstraní první vrstvu uvozovek a text pouze vypíše.

Proto je v definici SWAP toto jmeno z bezpečnostních důvodů obalené.
Jinak hrozí například při duplikuci definice `SWAP`, že podruhé bude název makra `SWAP` nahrazen za jeho obsah.

- `ex   DE` by se stal novým názvem makra a obsahoval by mezeru 
- mezera v názvu by zabránila nalezení makra, protože M4 by hledal zda `ex` není makro a pak to samé u `DE`
- za neočekávanou čárkou by znaky `HL` začínal druhý parametr
- původně druhý parametr by byl až třetí
- makro `define` by selhalo na přebytečnou čárku a díky ní neočekávaný třetí parametr

Proto musí být i čárka v původní definici `ex DE, HL` obalena.

Protože standartní M4 uvozovky jsou dost nepraktické a znemožňují použití apostrofů v textu, tak je na začátku předefinovávám za znaky `{` a `}`.

    changequote({,})dnl

Ty jsou doufám méně časté v bežném textu a navíc mě editor zvýrazňuje vzájemné párování. 
Navíc díky logickému párování jdou použít i v textu pokud splní dvě podmínky. 
Nikdy se uzavírací složená závorka nespáruje s otvírací řetězce a neukočí tak předčasně uvozený text.
A zároveň počet otvíracích složených závorek je shodný se zavíracími, takže není zbytek kodu spolknut jako parametr.

[Zpět na obsah](#obsah)

### 3. Nebezpečné znaky v parametrech makra

Jsou všechny znaky používané pro definici makra. 

- neuvozený název makra nebo rezervované slovo M4
- neuvozená kombinace `;#` kterou mám nastavenou pro M4 komentář do konce řádku
- neuvozený znak `,` neboli čárka je chápán jako oddělovač parametrů. Čárka jde ale obalit i do kulatých závorek.
- neuvozený znak `(` neboli otvírací závorka by se mohla v nepárovém provedení spárovat s uzavírací závorkou konce parametrů a vstup makra by přetekl do zbytku textu
- neuvozený znak `)` neboli uzavírací závorka by se mohla spárovat s otvírací závorkou parametrů a tím předčasne uzavřít makro
- to samé co platí pro závorky platí i pro nové znaky uvozovek `{` a `}`. Ty ale nejde nijak obalit, pouze splnit podmínku správného párování (viz předchozí kapitola) a nebo zapsat jako ascii hodnotu.

Hlavní nebezpečí teda číhá v makrech očekávající řetězce jako parametr. 
Například `PRINT({string})` kde navíc potřebujeme obalit string ještě jednou do uvozovek na `PRINT({"string"})`.
Znak `"` nejde prakticky použít v M4 jako uvozovky, protože by nešlo zanořovat obalení. Slouží ale jako uvození řetězce pro Pasmo compiler.

    PRINT("Bezpecny retezec")
    PRINT("Bezpecny retezec se spravne zarovnanymi zavorkami: (5+10)*2 = 30")
    PRINT({"Uvozeny nebezpecny retezec s dvojici znaku ;#"})
    PRINT({"Uvozeny nebezpecny retezec s nespravne zarovanymi zavorkami: )("})
    PRINT({"Uvozeny nebezpecny retezec obsahujici carku", 0x0D})
    PRINT({"Uvozeny nebezpecny retezec obsahujici spravne zarovnane uvozovky i={0..30}"})
    PRINT({"Uvozeny nebezpecny retezec obsahujici carku a neparovy uzaviraci slozeny znak ", 0x7D})

~~~    
./check_word.sh 'PRINT({"Retezec s jmenem makra SWAP a obsahujici specialni znaky (i+10+5)=i+15 a i={0..30}", 0x0D})'
    push DE             ; 1:11      print     
    ld   BC, size101    ; 3:10      print     Length of string101
    ld   DE, string101  ; 3:10      print     Address of string101
    call 0x203C         ; 3:17      print     Print our string with ZX 48K ROM
    pop  DE             ; 1:10      print

STRING_SECTION:
string101:
    db "Retezec s jmenem makra SWAP a obsahujici specialni znaky (i+10+5)=i+15 a i={0..30}", 0x0D
size101              EQU $ - string101
; seconds: 0           ;[11:58]
~~~
    
[Zpět na obsah](#obsah)

### 4. Odsazení v M4

Předchozí definice makra `SWAP` měla jeden vedlejší efekt.
Na výstupu se přidal jeden prázdný řádek, protože znaky za uzavírací závorkou jako je konec řádku bude zkopírován na výstup.
Pro zrušení konce řádku použijeme makro `dnl` které odmaže vše co za ním následuje.

    define({SWAP}, {
        ex   DE, HL         ; 1:4       swap   ( b a -- a b )})dnl

Bez `dnl` M4 vše zpracuje a nebo vypíše. Jedinou vyjímkou jsou ignorované počáteční mezery v seznamu parametrů. 
V ukázce je to jediná mezera za `{SWAP},`.
Druhý parametr obsahuje zalomení řádku následované mezerama na začátku a M4 je neignoruje jen díky tomu, že jsou obalené do `{}`.
Znak `{` mu říká tady začíná druhý paremetr a ne až slovo `ex`.

Abych mohl odsazovat C stylem ve zdrojáku řádky a nepřidával zbytečně mezery na výstup používám makro s názvem `__` (dvě podtržítka), které je definované tak že má prázdný výstup.
Makra místo mezery jdou od sebe oddělit pomoci kulatých závorek s prázdným parametrem.

    __()__()Odsazeny text ve zdrojaku, ale uz ne na vystupu

Případně jde použít i znaky uvozovek pro začátek a konec.

    __{}__{}Odsazeny text ve zdrojaku, ale uz ne na vystupu

Definice makra `__` vypadá takto:

    define({__},{}){}dnl

Stejného efektu by šlo dosáhnout i přes:

    define(   {__},                        ){}dnl

M4 obsahuje standartní znak `#` pro komentář. Ten na začátku měním za dvojznak `;#` a tím si uvolňuji celkem běžný znak #, používaný pro hex zápis čísla.
Komentář má podobný efekt jako makro `dnl` s tím rozdílem, že co za ním následuje je vypsáno na výstup, ale nezpracováno.

[Zpět na obsah](#obsah)

### 5. Definice maker v M4 

Na začátku první makra vypadala nějak takto:

    dnl
    define({SWAP},{
        ex   DE, HL         ; 1:4       swap   ( b a -- a b )}){}dnl
    dnl
    define({SUB},{        
        ex   DE, HL         ; 1:4       -   ( x1 x0 -- x )  x == x1 - x0
        or    A             ; 1:4       -
        sbc  HL, DE         ; 2:15      -
        pop  DE             ; 1:10      -}){}dnl
    dnl
        
Takže pokud na vstupu bylo: 

    SWAP SUB

Na výstup se zobrazilo:

        ex   DE, HL         ; 1:4       swap   ( b a -- a b )
        ex   DE, HL         ; 1:4       -   ( x1 x0 -- x )  x == x1 - x0
        or    A             ; 1:4       -
        sbc  HL, DE         ; 2:15      -
        pop  DE             ; 1:10      -

To ukazuje první bolest překladačů. 
Makra si nepředávají žádné informace mezi sebou, takže se duplikuje `ex DE,HL`. 
Proto na místech, kde mě to vadilo nejvíce jsem vytvořil nové slova.
Něco co bych přirovnal k optimalizaci kukátka (peephole optimization).

    dnl
    define({SWAP_SUB},{
        or    A             ; 1:4       swap -   ( x1 x0 -- x )  x == x0 - x1
        sbc  HL, DE         ; 2:15      swap -
        pop  DE             ; 1:10      swap -}){}dnl
    dnl

Pokud na vstup zapíšeme:

    SWAP_SUB
    
Tak se nyní zobrazí:

        or    A             ; 1:4       swap -   ( x1 x0 -- x )  x == x0 - x1
        sbc  HL, DE         ; 2:15      swap -
        pop  DE             ; 1:10      swap -
        
To má ale jednu nevýhodu.
Aby to šlo pužít musíte znát všechny vytvořené kombinace. 

[Zpět na obsah](#obsah)

### 6. Tokenizace a pole __TOKEN[] 

Aby se nemuseli znát všechny kombinovaná slova, tak jsem vytvořil shell script, který sám spojoval slova `SWAP` a `SUB` do kombinovanych slov `SWAP_SUB`. 
Ten se musel spouštět těsně před vlastní kompilací.

Později jsem našel způsob jak se bez skriptu obejít a přímo tokenizovat slova v M4 a spojení udělat až v tokenových pravidlech.

    dnl
    define({SWAP},{dnl
    __{}__ADD_TOKEN({__TOKEN_SWAP},{swap},$@){}dnl
    }){}dnl
    dnl
    define({SUB},{dnl        
    __{}__ADD_TOKEN({__TOKEN_SUB},{-},$@){}dnl
    }){}dnl
    dnl

Místo přímého výstupu v instrukcích makra postupně volají `__ADD_TOKEN`, který vytváří tokeny. 
Ukladá je do pole struktur `__TOKEN`, která vypadá takto:

    __TOKEN[id].NAME
    __TOKEN[id].INFO
    __TOKEN[id].PARAM
    
Ve skutečnosti to není pro M4 struktura, ale tři nezávislá makra s neplatnými názvy.
Není to ani pole, protože `id` může být obecně jakýkoliv řetezec.
Z mého pohledu je to pole, protože pro `id` používám postupně kladná čísla od 1 do počtu tokenů. 
Poslední použité `id` je uloženo v proměnné `__COUNT_TOKEN`, která je vlastně jen makro nahrazující svoje jméno za číslo.
Díky znakům `[`, `]` a `.` není celý řetezec standartní název makra a proto nebude nikdy nalezen a nahrazen za svůj obsah.

Abych spustil makro obsahující normálně nepovolené znaky musím ho volat přes pomocné makro `defn`

    defn({__TOKEN[}__COUNT_TOKEN{].NAME})

Protože ale chci z `__COUNT_TOKEN` získat číslo tak nesmí být makro uvozeno. 
Dalším způsobem by bylo zadat číslo přes (například zde první) parametr makra. Protože $1 se zamění vždy i kdyby bylo několikanásobně uvozené. 

    defn({__TOKEN[$1].NAME})

[Zpět na obsah](#obsah)

### 7. Pomocná makra pro práci s polem __TOKEN[] 

Pro jednodušší zápis pro přístup k datům jsem vytvořil pomocné makro `__T_NAME`

    define({__T_NAME},   {ifelse(eval(__COUNT_TOKEN>$1), 1,  {defn({__TOKEN[}eval(__COUNT_TOKEN-$1){].NAME})})}){}dnl

- `__T_NAME(0)` vrací název posledního uloženého tokenu pokud existuje a nebo prázdný řetezec
- `__T_NAME(1)` to samé pro předposlední token, atd.
- `__T_INFO(num)` vrací info pokud existuje a nebo prázdný řetezec 
- `__T_PARAM(num)` vrací parametry oddělené čarkama pokud existují a nebo prazdný řetezec a to vždy celé v závorce
- `__T_ARRAY(num)` to samé, ale bez uzavření do závorek
- `__T_ITEMS(num)` vrací nulu a nebo počet parametrů
- `__T_ARRAY_1(num)` pokud existuje vrátí první parametr
- `__T_ARRAY_2(num)` pokud existuje vrátí druhý parametr, atd. 
- `__T_REVERSE_1(num)` pokud existuje vrátí poslední parametr
- `__T_REVERSE_2(num)` pokud existuje vrátí předposlední parametr, atd.
- `__T_HEX_1(num)` pokud existuje a je to číslo vrátí první parametr ve tvaru 0x0000
- `__T_HEX_2(num)` pokud existuje a je to číslo vrátí druhý parametr ve tvaru 0x0000, atd. 
- `__T_HEX_REVERSE_1(num)` pokud existuje a je to číslo vrátí poslední parametr ve tvaru 0x0000
- `__T_HEX_REVERSE_2(num)` pokud existuje a je to číslo vrátí předposlední parametr ve tvaru 0x0000, atd.
- a mnohé další
    
V našem příkladu `SWAP SUB` makro `__ADD_TOKEN(__TOKEN_SWAP,{swap},$@)` prohledá tokenové pravidla.
Nic nenajde tak inkrementuje `__COUNT_TOKEN` na 1 a vytvoří makra

    __TOKEN[1].NAME=__TOKEN_SWAP
    __TOKEN[1].INFO={swap}
    __TOKEN[1].PARAM=()

Následující volání `__ADD_TOKEN(__TOKEN_SUB,{-},$@)` prohledá tokenové pravidla.
Zjistí že, existuje kombinace `SWAP_SUB` a místo uložení nového tokenu přepíše poslední na:

    __TOKEN[1].NAME=__TOKEN_SWAP_SUB
    __TOKEN[1].INFO={swap -}
    __TOKEN[1].PARAM=()

Na konci zpracování až proběhne celý vstupní soubor se zavolá makro `COMPILE()`, které až teď začne postupně procházet uložené tokeny a volat pro ně makra která je mění na instrukce Z80.
Tyhle makra se už nemohou jmenovat `SWAP` a `SUB`, protože ty existují a vytvářejí tokeny.
Takže se jmenují `__ASM_` + `_T_NAME(0)`.

    dnl
    define({__ASM_TOKEN_SWAP},{
        ex   DE, HL         ; 1:4       swap   ( b a -- a b )}){}dnl
    dnl
    define({__ASM_TOKEN_SUB},{        
        ex   DE, HL         ; 1:4       -   ( x1 x0 -- x )  x == x1 - x0
        or    A             ; 1:4       -
        sbc  HL, DE         ; 2:15      -
        pop  DE             ; 1:10      -}){}dnl
    dnl
    dnl
    define({__ASM_TOKEN_SWAP_SUB},{
        or    A             ; 1:4       swap -   ( x1 x0 -- x )  x == x0 - x1
        sbc  HL, DE         ; 2:15      swap -
        pop  DE             ; 1:10      swap -}){}dnl
    dnl
    
Což jsou původní verze maker jen pod jinými názvy. 

[Zpět na obsah](#obsah)

### 8. Tokenová pravidla 

Jsou uložena v souboru [__t_rules.m4](./M4/__t_rules.m4), kde `__ADD_TOKEN` začíná takto:
~~~
define({__ADD_TOKEN},{dnl
ifelse(__VERBOSE_LEVEL,1,{errprint({   ...new }$1{(}shift(shift($@)){) "}$2{"})}){}dnl
ifdef({__COUNT_TOKEN},
    {ifelse(dnl
= <> etc.,,,
    __T_NAME(0):$1, __TOKEN_SWAP:__TOKEN_SUB, {__SET_TOKEN({__TOKEN_SWAP_SUB},__CONCATENATE_WITH({ },__T_INFO(0),$2))},

    __T_NAME(0):$1, __TOKEN_SWAP:__TOKEN_CEQ, {__SET_TOKEN({__TOKEN_CEQ},__CONCATENATE_WITH({ },__T_INFO(0),$2))},
    __T_NAME(0):$1, __TOKEN_SWAP:__TOKEN_CNE, {__SET_TOKEN({__TOKEN_CNE},__CONCATENATE_WITH({ },__T_INFO(0),$2))},
    __T_NAME(0):$1, __TOKEN_SWAP:__TOKEN_HEQ, {__SET_TOKEN({__TOKEN_HEQ},__CONCATENATE_WITH({ },__T_INFO(0),$2))},
    __T_NAME(0):$1, __TOKEN_SWAP:__TOKEN_HNE, {__SET_TOKEN({__TOKEN_HNE},__CONCATENATE_WITH({ },__T_INFO(0),$2))},
...
~~~

Podmínka s `__VERBOSE_LEVEL` je pro podrobný debugující vypis.
Následuje podmínka zda je `__COUNT_TOKEN` definován a pokud ano následuje masivní `ifelse` větvení pro jednotlivé kombinace.
Na pořadí záleží, protože po první splněné podmínce se ostatní nesplní i kdyby platily.

#### Poznámka k formátování
Tahle část je zapsána právě takto
~~~
dnl
= <> etc.,,,
~~~
protože to `dnl` ukončí ignorované bílé znaky před začátkem parametru a první znak dalšího řádku je brán jako parametr.
Platný zápis by byl:
~~~
    {ifelse(
    __T_NAME(0):$1, __TOKEN_SWAP:__TOKEN_SUB, {__SET_TOKEN({__TOKEN_SWAP_SUB},__CONCATENATE_WITH({ },__T_INFO(0),$2))},
~~~
nebo
~~~
    {ifelse(dnl
__{}__T_NAME(0):$1, __TOKEN_SWAP:__TOKEN_SUB, {__SET_TOKEN({__TOKEN_SWAP_SUB},__CONCATENATE_WITH({ },__T_INFO(0),$2))},
~~~
Dále není možnost mezi podmínky vložit komentář, aby se nestalo to samé.
Takže komentář je vlastně zapsaný jako testovaný parametr a je porovnavan s prazdným řetezcem, který je mezi první a druhou čárkou.
Následuje prázdný řetezec ležící mezi druhou a třetí čárkou, který se aktivuje, kdyby se první dva parametry shodovaly.
Za třetí čárkou uz leží opět ignorované mezery pro další porovnání.

Pro náš příklad nás zajímá tenhle řádek
~~~
    __T_NAME(0):$1, __TOKEN_SWAP:__TOKEN_SUB, {__SET_TOKEN({__TOKEN_SWAP_SUB},__CONCATENATE_WITH({ },__T_INFO(0),$2))},
~~~
kde se porovnává zda předchozí token není `__TOKEN_SWAP` a právě vkládaný není `__TOKEN_SUB`.
Pokud ano tak se volá `__SET_TOKEN` který je definován takto:
~~~
define({__SET_TOKEN},{dnl
__{}define({__TOKEN[}__COUNT_TOKEN{].NAME},{$1}){}dnl
__{}define({__TOKEN[}__COUNT_TOKEN{].INFO},{$2}){}dnl
__{}define({__TOKEN[}__COUNT_TOKEN{].PARAM},(shift(shift($@)))){}dnl
}){}dnl
~~~
První parametr uloží do aktuálního tokenu jako `NAME`.
Přepíše ho, protože nemění `__COUNT_TOKEN`.
Druhý parametr uloží jako `INFO`.
Pak odmaže pomočí dvojitého shiftu první dva parametry a co zůstane uloží v závorce jako `PARAM`.
V našem příkladě volámne `__SET_TOKEN` jen s dvěma parametry, takže to uloží jen prázné závorky.

Druhý parametr by šel snadněji zadat jako `__T_INFO(0){ }$2`.
Místo toho volám makro `__CONCATENATE_WITH()`, které vrací řetezec ze spojených parametrů, kde čím se spojí určuje první parametr.
Je to z toho důvodu že občas může mít info hodnotu `__dtto`, které značí že je shodné s předchozím pokud se s ním spojí tak se má ignorovat.
Občas se nespojí a pak se nahradí za to předchozí info. 
Takže navenek to vypadá jako jeden celek se shodným infem, ale jsou to vlastně dvě makra.

Někdy je lepší Forth slovo rozdělit na dvě části, u kterých je šance že se budou lépe kombinovat s okolím.
Nejlépe to funguje když výsledný kód obsahuje instrukci `ex DE,HL`, která je identická se slovem `SWAP`.
Pak je výhodné se pokusit v kódu přesunout `ex DE,HL` na začátek nebo konec a vyjmout ho ven.
Například slovo `_2SWAP` vytváří tokeny pro `SWAP` následované tokenem `SWAP_2SWAP_SWAP` a ukončené tokenem `SWAP`.
~~~
./check_word.sh '_2SWAP'
    ex   DE, HL         ; 1:4       2swap   ( b a -- a b )
    ld    C, E          ; 1:4       2swap   ( d c b a -- a b c d )
    ld    B, D          ; 1:4       2swap   d c . b a     BC = b
    pop  DE             ; 1:10      2swap   d   . c a
    ex  [SP],HL         ; 1:19      2swap   a   . c d
    push BC             ; 1:11      2swap   a b . c d
    ex   DE, HL         ; 1:4       2swap   ( b a -- a b )
; seconds: 1           ;[ 7:56]
~~~
To pak umožnuje odstranit `SWAP` přes optimalizovaná slova jako je například `PUSH_SWAP()` nebo `SWAP_PUSH()`, aniž by se musela vytvářet další slova jako `PUSH_2SWAP()` nebo `_2SWAP_PUSH()` a nebo `PUSH_2SWAP_PUSH()` a nebo dokonce `PUSH_2SWAP_PUSHS()`.
Nevýhodou spojených slov je, že je nemusíme najít v tokenových pravidlech, protože jsou tvořeny kombinací jiných tokenů, nebo mají vnitřně jiné jméno.
Musíme se prvně kouknout jaké tokeny dělají.

[Zpět na obsah](#obsah)

### 9. Předávané info v makrech 

Makra vyžadující parametry mají místo `swap -` složitější info, které se může navíc měnit.
Například pokud makro očekává jediný číselný parametr, ale předtím proběhlo `10 5 +`, které překladač díky známým konstantám převede na `15`, tak se `10 5 +` ve výpise ztratí.
Aby se zachovala historie, tak se použije aktualní `__T_INFO(aktualni)` uložené do `__COMPILE_INFO`.

    dnl
    define({__ASM_TOKEN_SWAP},{
    __{}define({__INFO},__COMPILE_INFO){}dnl
        ex   DE, HL         ; 1:4       __INFO   ( b a -- a b )}){}dnl
    dnl
    define({__ASM_TOKEN_SUB},{
    __{}define({__INFO},__COMPILE_INFO){}dnl
        ex   DE, HL         ; 1:4       __INFO   ( x1 x0 -- x )  x == x1 - x0
        or    A             ; 1:4       __INFO
        sbc  HL, DE         ; 2:15      __INFO
        pop  DE             ; 1:10      __INFO}){}dnl
    dnl
    dnl
    define({__ASM_TOKEN_SWAP_SUB},{
    __{}define({__INFO},__COMPILE_INFO){}dnl
        or    A             ; 1:4       __INFO   ( x1 x0 -- x )  x == x0 - x1
        sbc  HL, DE         ; 2:15      __INFO
        pop  DE             ; 1:10      __INFO}){}dnl
    dnl
    
~~~
./check_word.sh 'PUSH(5) PUSH(10) ADD SUB'
    ld   BC, 0xFFF1     ; 3:10      5 10 + -   ( x -- x-0x000F )
    add  HL, BC         ; 1:11      5 10 + -
~~~
Umožnuje to také na pozadí používat uplně jiný token než se zdá, ale se shodným výsledkem. 
Například spojení `SWAP OVER` ve skutečnosti volá slovo `TUCK`.
~~~
./check_word.sh 'VERBOSE(1) SWAP OVER'
   ...new __TOKEN_SWAP() "swap" --> token(1) __TOKEN_SWAP "swap"
   ...new __TOKEN_OVER() "over" --> token(1) __TOKEN_TUCK "swap over"

   ...check all tokens
   ...second pass(1) __TOKEN_TUCK() "swap over" --> same

   ...check all tokens2
   ...third pass(1) __TOKEN_TUCK() "swap over" --> same

   ...create(1) __TOKEN_TUCK "swap over"
    push HL             ; 1:11      swap over   ( b a -- a b a )
; seconds: 0           ;[ 1:11]
~~~

[Zpět na obsah](#obsah)

### 10. Debugování 

- Uložené tokeny jdou zobrazit pomocí makra `__SHOW_TOKEN(id)`
~~~
./check_word.sh 'SWAP SUB __SHOW_TOKEN(1)'
; name: __TOKEN_SWAP_SUB
; info: swap -
;items: 0
;param: ()
    or    A             ; 1:4       swap -   ( x1 x0 -- x )  x == x0 - x1
    sbc  HL, DE         ; 2:15      swap -
    pop  DE             ; 1:10      swap -
; seconds: 0           ;[ 4:29]
~~~
- Celý proces tokenizace jde zobrazit přes makro `VERBOSE(1)`
~~~
./check_word.sh 'VERBOSE(1) SWAP SUB'
   ...new __TOKEN_SWAP() "swap" --> token(1) __TOKEN_SWAP "swap"
   ...new __TOKEN_SUB() "-" --> token(1) __TOKEN_SWAP_SUB "swap -"

   ...check all tokens
   ...second pass(1) __TOKEN_SWAP_SUB() "swap -" --> same

   ...check all tokens2
   ...third pass(1) __TOKEN_SWAP_SUB() "swap -" --> same

   ...create(1) __TOKEN_SWAP_SUB "swap -"
    or    A             ; 1:4       swap -   ( x1 x0 -- x )  x == x0 - x1
    sbc  HL, DE         ; 2:15      swap -
    pop  DE             ; 1:10      swap -
; seconds: 0           ;[ 4:29]
~~~
- Speciální informace o smyčkách jdou zobrazit přes makro `__SHOW_LOOP(id)`
~~~
./check_word.sh 'PUSH(10,0) DO LOOP __SHOW_LOOP(101)'
;      id: 101
;    type: M   ;# M,R,S = memory, recursive, stack
;  typ do: 
;   8-bit: 1
;     end: 10
;real end: 10
;   begin: 0
;    step: 1
    ld    A, 0x00       ; 2:7       10 0 do_101(xm)   8-bit loop
do101saveA:             ;           10 0 do_101(xm)
    ld  [idx101],A      ; 3:13      10 0 do_101(xm)
                        ;[9:45/32]  loop_101(xm)   variant +1.nop: 8-bit loop and hi(index)=0, run 10x
idx101 EQU $+1          ;           loop_101(xm)   idx always points to a 16-bit index
    ld    A, 0          ; 2:7       loop_101(xm)   0.. +1 ..(10), real_stop:0x000A
    nop                 ; 1:4       loop_101(xm)   hi(index) = 0 = nop -> idx always points to a 16-bit index.
    inc   A             ; 1:4       loop_101(xm)   index++
    cp   0x0A           ; 2:7       loop_101(xm)   lo(real_stop)
    jp   nz, do101saveA ; 3:10      loop_101(xm)   index-real_stop
leave101:               ;           loop_101(xm)
exit101:                ;           loop_101(xm)
; seconds: 0           ;[14:52]
~~~
- M4 obsahuje rezervované slovo `dumpdef({macro_name})`, které zobrazí kód makra
~~~
./check_word.sh 'dumpdef({SWAP})'
SWAP:	dnl
__{}__ADD_TOKEN({__TOKEN_SWAP},{swap},$@){}dnl

; seconds: 0           ;[ 0:0]
~~~
- Protože u `dumpdef` bez znalosti tvoření jmen uvidíme jen tvoření tokenu tak jsem udělal `DEBUG({word})`
~~~
./check_word.sh 'DEBUG({SWAP})'
**************** head (create token) ****************
SWAP:	dnl
__{}__ADD_TOKEN({__TOKEN_SWAP},{swap},$@){}dnl

**************** body (make asm) ****************
__ASM_TOKEN_SWAP:	dnl
__{}define({__INFO},__COMPILE_INFO)
    ex   DE, HL         ; 1:4       __INFO   ( b a -- a b )
; seconds: 0           ;[ 0:0]
~~~
- Někdy nevíme v kterém připojeném definičním souboru maker je chyba a zobrazují se nechtěné znaky. Pro podrobnější info používám přídání znaku a půlení intervalu.
~~~
../check_word.sh --check 'SWAP SUB'
;# vvv --- __macros.m4
;# vvv --- __eval.m4
;# vvv --- __t_rules.m4
;# vvv --- float.m4
;# vvv --- if.m4
;# vvv --- case.m4
;# vvv --- logic.m4
;# vvv --- function.m4
;# vvv --- loop.m4
;# vvv --- device.m4
;# vvv --- arithmetic.m4
;# vvv --- memory.m4
;# vvv --- other.m4
;# vvv --- stack.m4
;# vvv --- array.m4
;# vvv --- zx48float.m4
    ;# ^^^ ---
 
    or    A             ; 1:4       swap -   ( x1 x0 -- x )  x == x0 - x1
    sbc  HL, DE         ; 2:15      swap -
    pop  DE             ; 1:10      swap - ;# vvv --- float_runtime.m4
;# vvv --- graphic_runtime.m4
;# vvv --- zx48float_runtime.m4
;# vvv --- runtime.m4
;# ^^^ ---
                       ;[ 4:29]
~~~

[Zpět na obsah](#obsah)
