#!/bin/sh

help_info () {
	echo "$0 is bash/awk script that converts the original Forth source program to M4 Forth." >&2
	echo "" >&2
	echo "   Use: $0 -h -zfloat source.fth > dest.m4 " >&2
	echo "" >&2
	echo "	if file start with '-' char, you can use -- before file1" >&2
	echo "" >&2
	echo "	 -zfloat ...switch floating point arithmetic from default 16-bit to 40-bit using ZX Spectrum ROM" >&2
	echo "	 -h      ...show this message" >&2
	echo "" >&2
}

zfloat=""
help=0
fail=0
only_file=0
file=""

AWK_SCRIPT="$(dirname "$(readlink -f "$0")")/fth2m4.awk"

#handle command line options
while [ $# -gt 0 ] ; do
	if [ "$only_file" = "1" ] && [ "$file" = "" ] ; then
		file=$1
	elif [ "$only_file" = "1" ] ; then
		echo "Error: Another source file '$1' will be ignored" >&2
		echo "" >&2
		help=1
		fail=1
	elif [ "$1" = "-zfloat" ] ; then
		zfloat=$1
	elif [ "$1" = "-h" ] ; then
		help=1
	elif [ "$1" = "--" ] ; then
		only_file=1
	elif [ "$file" = "" ] ; then
		file=$1
	else
		echo "Error: Another source file '$1' will be ignored" >&2
		echo "" >&2
		help=1
		fail=1
	fi
	shift
done

if [ "$file" = "" ] ; then
   [ "$help" = "0" ] && printf "Error: Need filename with forth!\n" >&2 && echo "" >&2
   help=1
   fail=1
elif [ ! -s "$AWK_SCRIPT" ] ; then
   printf "Error: \"$AWK_SCRIPT\" not found or empty!\n" >&2 && echo "" >&2
   fail=1
elif [ ! -s "$file" ] ; then
   printf "Error: \"$file\" not found or empty!\n" >&2 && echo "" >&2
   help=1
   fail=1
fi


[ "$help" = "1" ] && help_info

[ "$fail" = "1" ] && exit 1

awk -v arg="$zfloat" -f "$AWK_SCRIPT" -- "$file"
